# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

import datetime
import decimal
import unittest
import warnings
from astm import mapping

class FieldTestCase(unittest.TestCase):

    def test_init_default(self):
        f = mapping.Field()
        self.assertTrue(hasattr(f, 'name'))
        self.assertTrue(hasattr(f, 'default'))
        self.assertEqual(f.name, None)
        self.assertEqual(f.default, None)

    def test_init_with_custom_name(self):
        f = mapping.Field(name='foo')
        self.assertEqual(f.name, 'foo')

    def test_init_with_custom_default_value(self):
        f = mapping.Field(default='foo')
        self.assertEqual(f.default, 'foo')

    def test_callable_default_value(self):
        class Dummy(mapping.Mapping):
            field = mapping.Field(default=lambda: 'foobar')
        self.assertEqual(Dummy().field, 'foobar')


class NotUsedFieldTestCase(unittest.TestCase):

    def setUp(self):
        class Dummy(mapping.Mapping):
            field = mapping.NotUsedField()
        self.Dummy = Dummy

    def test_get_value(self):
        obj = self.Dummy()
        self.assertEqual(obj.field, None)
        self.assertEqual(obj[0], None)

    def test_set_value(self):
        obj = self.Dummy()
        with warnings.catch_warnings(record=True) as w:
            obj.field = 42
            assert issubclass(w[-1].category, UserWarning)
        self.assertEqual(obj.field, None)


class IntegerTestCase(unittest.TestCase):

    def setUp(self):
        class Dummy(mapping.Mapping):
            field = mapping.IntegerField()
        self.Dummy = Dummy

    def test_get_value(self):
        obj = self.Dummy(field=42)
        self.assertEqual(obj.field, 42)

    def test_set_value(self):
        obj = self.Dummy()
        obj.field = 42
        self.assertEqual(obj.field, 42)

    def test_set_string_value(self):
        obj = self.Dummy()
        obj.field = '42'
        self.assertEqual(obj.field, 42)
        self.assertRaises(TypeError, setattr, obj, 'field', 'foo')


class DecimalFieldTestCase(unittest.TestCase):

    def setUp(self):
        class Dummy(mapping.Mapping):
            field = mapping.DecimalField()
        self.Dummy = Dummy

    def test_get_value(self):
        obj = self.Dummy(field=3.14)
        self.assertEqual(obj.field, decimal.Decimal('3.14'))

    def test_set_value(self):
        obj = self.Dummy()
        obj.field = 3.14
        self.assertEqual(obj.field, decimal.Decimal('3.14'))

    def test_set_int_value(self):
        obj = self.Dummy()
        obj.field = 42
        self.assertEqual(obj.field, decimal.Decimal('42'))


class TextFieldTestCase(unittest.TestCase):

    def setUp(self):
        class Dummy(mapping.Mapping):
            field = mapping.TextField()
        self.Dummy = Dummy

    def test_get_value(self):
        obj = self.Dummy(field='foo')
        self.assertEqual(obj.field, 'foo')

    def test_set_value(self):
        obj = self.Dummy()
        obj.field = 'привет'
        self.assertEqual(obj.field, 'привет')

    def test_set_utf8_value(self):
        obj = self.Dummy()
        obj.field = 'привет'.encode('utf-8')
        self.assertEqual(obj.field, 'привет')

    def test_fail_set_non_utf8_value(self):
        obj = self.Dummy()
        try:
            obj.field = 'привет'.encode('cp1251')
        except UnicodeDecodeError:
            pass
        else:
            self.fail('%s expected' % UnicodeDecodeError)

    def test_fail_set_non_string_value(self):
        obj = self.Dummy()
        try:
            obj.field = object()
        except TypeError:
            pass
        else:
            self.fail('%s expected' % TypeError)

    def test_raw_value(self):
        obj = self.Dummy()
        obj.field = 'привет'
        self.assertEqual(obj._data['field'], 'привет')


class DateFieldTestCase(unittest.TestCase):

    def setUp(self):
        class Dummy(mapping.Mapping):
            field = mapping.DateField()
        self.Dummy = Dummy
        self.datetime = datetime.datetime(2009, 2, 13, 23, 31, 30)
        self.date = datetime.datetime(2009, 2, 13)

    def test_get_value(self):
        obj = self.Dummy(field=self.datetime)
        self.assertEqual(obj.field, self.date)

    def test_set_datetime_value(self):
        obj = self.Dummy()
        obj.field = self.datetime
        self.assertEqual(obj.field, self.date)

    def test_init_date_value(self):
        obj = self.Dummy(field=self.date)
        self.assertEqual(obj.field, self.date)

    def test_set_date_value(self):
        obj = self.Dummy()
        obj.field = self.date
        self.assertEqual(obj.field, self.date)

    def test_raw_value(self):
        obj = self.Dummy()
        obj.field = self.datetime
        self.assertEqual(obj._data['field'],
                         self.date.strftime(obj._fields[0][1].format))

    def test_set_string_value(self):
        obj = self.Dummy()
        obj.field = '20090213'
        self.assertRaises(ValueError, setattr, obj, 'field', '1234567')


class TimeFieldTestCase(unittest.TestCase):

    def setUp(self):
        class Dummy(mapping.Mapping):
            field = mapping.TimeField()
        self.Dummy = Dummy
        self.datetime = datetime.datetime(2009, 2, 13, 23, 31, 30)
        self.time = datetime.time(23, 31, 30)

    def test_get_value(self):
        obj = self.Dummy(field=self.datetime)
        self.assertEqual(obj.field, self.time)

    def test_set_datetime_value(self):
        obj = self.Dummy()
        obj.field = self.datetime
        self.assertEqual(obj.field, self.time)

    def test_init_time_value(self):
        obj = self.Dummy(field=self.time)
        self.assertEqual(obj.field, self.time)

    def test_set_time_value(self):
        obj = self.Dummy()
        obj.field = self.time
        self.assertEqual(obj.field, self.time)

    def test_raw_value(self):
        obj = self.Dummy()
        obj.field = self.datetime
        self.assertEqual(obj._data['field'],
                         self.time.strftime(obj._fields[0][1].format))

    def test_set_string_value(self):
        obj = self.Dummy()
        obj.field = '111213'
        self.assertRaises(ValueError, setattr, obj, 'field', '314159')


class DatetimeFieldTestCase(unittest.TestCase):

    def setUp(self):
        class Dummy(mapping.Mapping):
            field = mapping.DateTimeField()
        self.Dummy = Dummy
        self.datetime = datetime.datetime(2009, 2, 13, 23, 31, 30)
        self.date = datetime.datetime(2009, 2, 13)

    def test_get_value(self):
        obj = self.Dummy(field=self.datetime)
        self.assertEqual(obj.field, self.datetime)

    def test_set_value(self):
        obj = self.Dummy()
        obj.field = self.datetime
        self.assertEqual(obj.field, self.datetime)

    def test_get_date_value(self):
        obj = self.Dummy(field=self.date)
        self.assertEqual(obj.field, self.date)

    def test_set_date_value(self):
        obj = self.Dummy()
        obj.field = self.date
        self.assertEqual(obj.field, self.date)

    def test_raw_value(self):
        obj = self.Dummy()
        obj.field = self.datetime
        self.assertEqual(obj._data['field'],
                         self.datetime.strftime(obj._fields[0][1].format))

    def test_set_string_value(self):
        obj = self.Dummy()
        obj.field = '20090213233130'
        self.assertRaises(ValueError, setattr, obj, 'field', '12345678901234')


class ConstantFieldTestCase(unittest.TestCase):

    def test_get_value(self):
        class Dummy(mapping.Mapping):
            field = mapping.ConstantField(default=42)
        obj = Dummy()
        self.assertEqual(obj.field, 42)

    def test_set_value_if_none_default(self):
        class Dummy(mapping.Mapping):
            field = mapping.ConstantField(default='foo')
        obj = Dummy()
        obj.field = 'foo'
        self.assertEqual(obj.field, 'foo')

    def test_fail_override_setted_value(self):
        class Dummy(mapping.Mapping):
            field = mapping.ConstantField(default='foo')
        obj = Dummy()
        obj.field = 'foo'
        self.assertEqual(obj.field, 'foo')
        self.assertRaises(ValueError, setattr, obj, 'field', 'bar')

    def test_restrict_new_values_by_default_one(self):
        class Dummy(mapping.Mapping):
            field = mapping.ConstantField(default='foo')
        obj = Dummy()
        self.assertRaises(ValueError, setattr, obj, 'field', 'bar')
        obj.field = 'foo'
        self.assertEqual(obj.field, 'foo')

    def test_raw_value(self):
        class Dummy(mapping.Mapping):
            field = mapping.ConstantField(default='foo')
        obj = Dummy()
        obj.field = 'foo'
        self.assertEqual(obj._data['field'], 'foo')

    def test_raw_value_should_be_string(self):
        class Dummy(mapping.Mapping):
            field = mapping.ConstantField(default=42)
        obj = Dummy()
        obj.field = 42
        self.assertEqual(obj._data['field'], '42')

    def test_always_required(self):
        field = mapping.ConstantField(default='test')
        assert field.required
        self.assertRaises(TypeError, mapping.ConstantField, default='test', required=False)

    def test_default_value_should_be_defined(self):
        self.assertRaises(ValueError, mapping.ConstantField)


class SetFieldTestCase(unittest.TestCase):

    def setUp(self):
        class Dummy(mapping.Mapping):
            field = mapping.SetField(values=['foo', 'bar', 'baz'])
        self.Dummy = Dummy

    def test_get_value(self):
        obj = self.Dummy(field='foo')
        self.assertEqual(obj.field, 'foo')

    def test_set_value(self):
        obj = self.Dummy()
        obj.field = 'bar'
        self.assertEqual(obj.field, 'bar')

    def test_restrict_new_values_by_specified_set(self):
        obj = self.Dummy()
        self.assertRaises(ValueError, setattr, obj, 'field', 'boo')

    def test_reject_any_value(self):
        class Dummy(mapping.Mapping):
            field = mapping.SetField()
        obj = Dummy()
        self.assertRaises(ValueError, setattr, obj, 'field', 'bar')
        self.assertRaises(ValueError, setattr, obj, 'field', 'foo')
        obj.field = None

    def test_custom_field(self):
        class Dummy(mapping.Mapping):
            field = mapping.SetField(values=[1, 2, 3],
                                     field=mapping.IntegerField())
        obj = Dummy()
        obj.field = 1
        self.assertEqual(obj._data['field'], '1')
        obj.field = 2
        self.assertEqual(obj._data['field'], '2')
        obj.field = 3
        self.assertEqual(obj._data['field'], '3')


    def test_raw_value(self):
        obj = self.Dummy()
        obj.field = 'foo'
        self.assertEqual(obj._data['field'], 'foo')


class ComponentFieldTestCase(unittest.TestCase):

    def setUp(self):
        class Dummy(mapping.Mapping):
            field = mapping.ComponentField(
                mapping = mapping.Component.build(
                    mapping.Field(name='foo'),
                    mapping.IntegerField(name='bar'),
                    mapping.ConstantField(name='baz', default='42')
                )
            )
        self.Dummy = Dummy

    def test_get_value(self):
        obj = self.Dummy(field=['foo', 14, '42'])
        self.assertEqual(obj.field, ['foo', 14, '42'])

    def test_set_value(self):
        obj = self.Dummy()
        self.assertRaises(TypeError, setattr, obj, 'field', 42)
        self.assertRaises(ValueError, setattr, obj, 'field', [1, 2, 3])
        obj.field = ['test', 24, '42']
        self.assertEqual(obj.field, ['test', 24, '42'])

    def test_iter(self):
        obj = self.Dummy(field=['foo', 14, '42'])
        self.assertEqual(list(obj.field), ['foo', 14, '42'])

    def test_raw_value(self):
        obj = self.Dummy()
        obj.field = ['foo', 14, '42']
        self.assertEqual(obj._data['field'], ['foo', 14, '42'])

    def test_set_string_value(self):
        obj = self.Dummy()
        obj.field = 'foo'
        self.assertEqual(obj.field[0], 'foo')
        self.assertEqual(obj.field[1], None)
        self.assertEqual(obj.field[2], '42')


class RepeatedComponentFieldTestCase(unittest.TestCase):

    def setUp(self):
        class Dummy(mapping.Mapping):
            field = mapping.RepeatedComponentField(
                mapping.Component.build(
                    mapping.TextField(name='key'),
                    mapping.IntegerField(name='value'),
                )
            )
        class Thing(mapping.Mapping):
            numbers = mapping.RepeatedComponentField(
                mapping.Component.build(
                    mapping.IntegerField(name='value')
                )
            )
        self.Dummy = Dummy
        self.Thing = Thing

    def test_get_value(self):
        obj = self.Dummy(field=[['foo', 1], ['bar', 2], ['baz', 3]])
        self.assertEqual(obj.field, [['foo', 1], ['bar', 2], ['baz', 3]])

    def test_set_value(self):
        obj = self.Dummy()
        self.assertRaises(TypeError, setattr, obj, 'field', 42)
        obj.field = [['foo', 42]]
        self.assertEqual(obj.field, [['foo', 42]])

    def test_fail_on_set_strings(self):
        obj = self.Dummy()
        obj.field = 'foo' # WHY?
        #self.assertRaises(TypeError, setattr, obj, 'field', 'foo')

    def test_iter(self):
        obj = self.Dummy(field=[['foo', 14]])
        self.assertEqual(list(obj.field), [['foo', 14]])

    def test_getter_returns_list(self):
        obj = self.Dummy([['foo', 42]])
        self.assertTrue(isinstance(obj.field, list))

    def test_proxy_delitem(self):
        obj = self.Dummy([['foo', 1], ['bar', 2]])
        del obj.field[0]
        self.assertEqual(len(obj.field), 1)
        self.assertEqual(obj.field[0], ['bar', 2])

    def test_proxy_append(self):
        obj = self.Dummy([['foo', 1]])
        self.assertEqual(obj.field[0], ['foo', 1])
        obj.field.append(['bar', 2])
        self.assertEqual(obj.field[1], ['bar', 2])

    def test_proxy_extend(self):
        obj = self.Dummy([['foo', 1]])
        obj.field.extend([['bar', 2], ['baz', 3]])
        self.assertEqual(len(obj.field), 3)
        self.assertEqual(obj.field[2], ['baz', 3])

    def test_proxy_contains(self):
        obj = self.Thing(numbers=[[i] for i in range(5)])
        self.assertTrue([3] in obj.numbers)
        self.assertTrue([6] not in obj.numbers)

    def test_proxy_count(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        self.assertEqual(1, obj.numbers.count([1]))
        self.assertEqual(0, obj.numbers.count([4]))

    def test_proxy_index(self):
        obj = self.Thing(numbers=[[1], [2], [4]])
        self.assertEqual(1, obj.numbers.index([2]))

    def test_proxy_index_range(self):
        obj = self.Thing(numbers=[[1], [2], [4], [5]])
        self.assertEqual(2, obj.numbers.index([4], 2, 3))

    def test_fail_proxy_index_for_nonexisted_element(self):
        obj = self.Thing(numbers=[[1], [2], [4]])
        self.assertRaises(ValueError, obj.numbers.index, [5])

    def test_fail_proxy_index_negative_start(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        self.assertRaises(ValueError, obj.numbers.index, 2, -1 ,3)

    def test_proxy_insert(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        obj.numbers.insert(0, [0])
        self.assertEqual(obj.numbers[0], [0])

    def test_proxy_remove(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        obj.numbers.remove([1])
        obj.numbers.remove([2])
        self.assertEqual(len(obj.numbers), 1)
        self.assertEqual(obj.numbers[0].value, 3)

    def test_fail_proxy_remove_missing(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        self.assertRaises(ValueError, obj.numbers.remove, [5])

    def test_proxy_pop(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        self.assertEqual(obj.numbers.pop(), [3])
        self.assertEqual(len(obj.numbers), 2)
        self.assertEqual(obj.numbers.pop(0), [1])

    def test_proxy_slices(self):
        obj = self.Thing()
        obj.numbers = [[i] for i in range(5)]
        ll = obj.numbers[1:3]
        self.assertEqual(len(ll), 2)
        self.assertEqual(ll[0], [1])
        obj.numbers[2:4] = [[i] for i in range(6, 8)]
        self.assertEqual(obj.numbers[2], [6])
        self.assertEqual(obj.numbers[4], [4])
        self.assertEqual(len(obj.numbers), 5)
        del obj.numbers[3:]
        self.assertEqual(len(obj.numbers), 3)

    def test_proxy_sort_fails(self):
        class Dummy(mapping.Mapping):
            numbers = mapping.RepeatedComponentField(
                mapping.Component.build(
                    mapping.IntegerField(name='a'),
                    mapping.IntegerField(name='b')
                )
            )
        obj = Dummy(numbers=[[4, 2], [2, 3], [0, 1]])
        self.assertRaises(NotImplementedError, obj.numbers.sort)

    def test_proxy_lt(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        self.assertTrue(obj.numbers < [[4], [5], [6]])

    def test_proxy_lt_with_other_proxy(self):
        obj1 = self.Thing(numbers=[[1], [2], [3]])
        obj2 = self.Thing(numbers=[[4], [5], [6]])
        self.assertTrue(obj1.numbers < obj2.numbers)

    def test_proxy_le(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        self.assertTrue(obj.numbers <= [[1], [2], [3]])

    def test_proxy_le_with_other_proxy(self):
        obj1 = self.Thing(numbers=[[1], [2], [3]])
        obj2 = self.Thing(numbers=[[1], [2], [3]])
        self.assertTrue(obj1.numbers <= obj2.numbers)

    def test_proxy_eq(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        self.assertTrue(obj.numbers == [[1], [2], [3]])

    def test_proxy_eq_with_other_proxy(self):
        obj1 = self.Thing(numbers=[[1], [2], [3]])
        obj2 = self.Thing(numbers=[[1], [2], [3]])
        self.assertTrue(obj1.numbers == obj2.numbers)

    def test_proxy_ne(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        self.assertTrue(obj.numbers != [[4], [5], [6]])

    def test_proxy_ne_with_other_proxy(self):
        obj1 = self.Thing(numbers=[[1], [2], [3]])
        obj2 = self.Thing(numbers=[[4], [5], [6]])
        self.assertTrue(obj1.numbers != obj2.numbers)

    def test_proxy_ge(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        self.assertTrue(obj.numbers >= [[1], [2], [3]])

    def test_proxy_ge_with_other_proxy(self):
        obj1 = self.Thing(numbers=[[1], [2], [3]])
        obj2 = self.Thing(numbers=[[1], [2], [3]])
        self.assertTrue(obj1.numbers >= obj2.numbers)

    def test_proxy_gt(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        self.assertTrue(obj.numbers > [[0], [1], [2]])

    def test_proxy_gt_with_other_proxy(self):
        obj1 = self.Thing(numbers=[[1], [2], [3]])
        obj2 = self.Thing(numbers=[[0], [1], [2]])
        self.assertTrue(obj1.numbers > obj2.numbers)

    def test_proxy_add(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        l = obj.numbers + [[4], [5], [6]]
        self.assertEqual(l, [[1], [2], [3], [4], [5], [6]])
        self.assertTrue(isinstance(l, list))
        self.assertTrue(obj.numbers is not l)

    def test_proxy_add_other_proxy(self):
        obj1 = self.Thing(numbers=[[1], [2], [3]])
        obj2 = self.Thing(numbers=[[4], [5], [6]])
        l = obj1.numbers + obj2.numbers
        self.assertEqual(l, [[1], [2], [3], [4], [5], [6]])

    def test_proxy_iadd(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        obj.numbers += [[4], [5], [6]]
        self.assertEqual(obj.numbers, [[1], [2], [3], [4], [5], [6]])

    def test_proxy_mul(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        l = obj.numbers * 2
        self.assertEqual(l, [[1], [2], [3], [1], [2], [3]])
        self.assertTrue(isinstance(l, list))
        self.assertTrue(obj.numbers is not l)
        self.assertEqual(obj.numbers, [[1], [2], [3]])

    def test_proxy_mul_one(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        l = obj.numbers * 1
        self.assertEqual(l, [[1], [2], [3]])

    def test_proxy_mul_zero(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        l = obj.numbers * 0
        self.assertEqual(l, [])
        self.assertTrue(isinstance(l, list))

    def test_proxy_imul(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        obj.numbers *= 2
        self.assertEqual(obj.numbers, [[1], [2], [3], [1], [2], [3]])

    def test_proxy_imul_one(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        obj.numbers *= 1
        self.assertEqual(obj.numbers, [[1], [2], [3]])

    def test_proxy_imul_zero(self):
        obj = self.Thing(numbers=[[1], [2], [3]])
        obj.numbers *= 0
        self.assertEqual(obj.numbers, [])


class MappingTestCase(unittest.TestCase):

    def setUp(self):
        class Dummy(mapping.Mapping):
            foo = mapping.Field(default='bar')
            bar = mapping.ComponentField(mapping=mapping.Component.build(
                mapping.IntegerField(name='a'),
                mapping.IntegerField(name='b'),
                mapping.IntegerField(name='c'),
            ), default=[1,2,3])
        class Thing(mapping.Mapping):
            numbers = mapping.RepeatedComponentField(
                mapping.Component.build(
                    mapping.IntegerField(name='a'),
                    mapping.IntegerField(name='b')
                )
            )
        self.Dummy = Dummy
        self.Thing = Thing

    def test_equal(self):
        obj = self.Dummy('foo', [3, 2, 1])
        self.assertEqual(obj, ['foo', (3, 2, 1)])
        self.assertNotEqual(obj, ['foo'])

    def test_iter(self):
        obj = self.Dummy('foo', [3, 2, 1])
        self.assertEqual(list(obj), ['foo', [3, 2, 1]])

    def test_len(self):
        obj = self.Dummy('foo', [3, 2, 1])
        self.assertEqual(len(obj), 2)
        self.assertEqual(len(obj.bar), 3)

    def test_contains(self):
        obj = self.Dummy('foo', [3, 2, 1])
        assert 'foo' in obj

    def test_getitem(self):
        obj = self.Dummy('foo', [3, 2, 1])
        self.assertEqual(obj[1][0], 3)

    def test_setitem(self):
        obj = self.Dummy('foo', [3, 2, 1])
        obj[1][0] = 42
        self.assertEqual(obj[1][0], 42)

    def test_delitem(self):
        obj = self.Dummy('foo', [3, 2, 1])
        del obj[1][1]
        self.assertEqual(obj[1][1], None)

    def test_to_astm_record(self):
        obj = self.Dummy('foo', [3, 2, 1])
        self.assertEqual(obj.to_astm(), ['foo', ['3', '2', '1']])
        obj = self.Thing(numbers=[[4, 2], [2, 3], [0, 1]])
        self.assertEqual(obj.to_astm(), [[['4', '2'], ['2', '3'], ['0', '1']]])

    def test_required_field(self):
        class Dummy(mapping.Mapping):
            field = mapping.Field(required=True)
        obj = Dummy()
        self.assertTrue(obj.field is None)
        self.assertRaises(ValueError, obj.to_astm)

    def test_field_max_length(self):
        class Dummy(mapping.Mapping):
            field = mapping.Field(length=10)
        obj = Dummy()
        obj.field = '-' * 9
        obj.field = None
        obj.field = '-' * 10
        self.assertRaises(ValueError, setattr, obj, 'field', '-' * 11)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#
import asyncio
import logging
from .codec import decode_message, is_chunked_message, join
from .constants import ACK, EOT, NAK, ENQ, ENCODING
from .exceptions import InvalidState, NotAccepted

log = logging.getLogger(__name__)

__all__ = ['BaseRecordsDispatcher', 'Server']


class BaseRecordsDispatcher(object):
    """Abstract dispatcher of received ASTM records.
    You need to override its handlers or extend the dispatcher for your needs.
    """
    encoding = ENCODING

    def __init__(self, encoding=None):
        self.encoding = encoding or self.encoding
        self.dispatch = {
            'H': self.on_header,
            'C': self.on_comment,
            'P': self.on_patient,
            'O': self.on_order,
            'R': self.on_result,
            'S': self.on_scientific,
            'M': self.on_manufacturer_info,
            'L': self.on_terminator
        }
        self.wrappers = {}

    def __call__(self, message):
        """
        Decodes and dispatches incoming message.

        :param message: Message to dispatch.
        :type message: bytes
        """
        seq, records, cs = decode_message(message, self.encoding)
        for record in records:
            handler = self.dispatch.get(record[0], self.on_unknown)
            handler(self.wrap(record))

    def wrap(self, record):
        """
        Wraps record to high-level object if wrapper is defined.
        
        :param record: ASTM record.
        :type record: list
        
        :return: High-level wrapper or raw record.
        """
        rtype = record[0]
        if rtype in self.wrappers:
            return self.wrappers[rtype](*record)
        return record

    def _default_handler(self, record):
        log.warning('Record remains unprocessed: %s', record)

    def on_header(self, record):
        """Header record handler."""
        self._default_handler(record)

    def on_comment(self, record):
        """Comment record handler."""
        self._default_handler(record)

    def on_patient(self, record):
        """Patient record handler."""
        self._default_handler(record)

    def on_order(self, record):
        """Order record handler."""
        self._default_handler(record)

    def on_result(self, record):
        """Result record handler."""
        self._default_handler(record)

    def on_scientific(self, record):
        """Scientific record handler."""
        self._default_handler(record)

    def on_manufacturer_info(self, record):
        """Manufacturer information record handler."""
        self._default_handler(record)

    def on_terminator(self, record):
        """Terminator record handler."""
        self._default_handler(record)

    def on_unknown(self, record):
        """Fallback handler for dispatcher."""
        self._default_handler(record)


async def handle_connection(reader, writer, dispatcher, encoding, timeout):
    """
    Handles single client connection.
    """
    chunks = []
    is_transfer_state = False
    peername = writer.get_extra_info('peername')
    log.info('Connection from %s', peername)

    async def read(n=1):
        try:
            return await asyncio.wait_for(reader.read(n), timeout)
        except asyncio.TimeoutError:
            log.warning('Connection timed out for %s', peername)
            writer.close()
            await writer.wait_closed()
            return None

    while True:
        data = await read()
        if not data:
            break

        if data == ENQ:
            if not is_transfer_state:
                is_transfer_state = True
                writer.write(ACK)
                await writer.drain()
            else:
                log.error('ENQ is not expected.')
                writer.write(NAK)
                await writer.drain()

        elif data == EOT:
            if is_transfer_state:
                is_transfer_state = False
            else:
                log.error('EOT is not expected.')
        
        elif data in (ACK, NAK):
            log.warning('%r is not expected on server side.', data)

        else: # Message frame
            if not is_transfer_state:
                log.error('Message frame is not expected.')
                writer.write(NAK)
                await writer.drain()
                continue
            
            frame = data + await reader.readuntil(b'\r')
            
            try:
                if is_chunked_message(frame):
                    chunks.append(frame)
                elif chunks:
                    chunks.append(frame)
                    dispatcher(join(chunks))
                    chunks = []
                else:
                    dispatcher(frame)
                writer.write(ACK)
                await writer.drain()
            except Exception:
                log.exception('Error handling message: %r', frame)
                writer.write(NAK)
                await writer.drain()

    log.info('Connection closed for %s', peername)


class Server:
    """
    Asynchronous ASTM server.

    :param host: Server IP address or hostname.
    :type host: str

    :param port: Server port number.
    :type port: int

    :param dispatcher: Custom request handler records dispatcher.
                       If omitted the :class:`BaseRecordsDispatcher` will be
                       used by default.
    :type dispatcher: :class:`BaseRecordsDispatcher`

    :param timeout: Connection timeout in seconds.
    :type timeout: int

    :param encoding: Dispatcher's encoding.
    :type encoding: str
    """
    dispatcher = BaseRecordsDispatcher

    def __init__(self, host='localhost', port=15200,
                 dispatcher=None, timeout=None, encoding=None):
        self.host = host
        self.port = port
        self.timeout = timeout
        self.encoding = encoding
        if dispatcher is not None:
            self.dispatcher = dispatcher
        self._server = None

    async def start(self):
        """Starts the server."""
        self._server = await asyncio.start_server(
            lambda r, w: handle_connection(
                r, w,
                self.dispatcher(encoding=self.encoding),
                self.encoding,
                self.timeout
            ),
            self.host,
            self.port
        )
        addrs = ', '.join(str(s.getsockname()) for s in self._server.sockets)
        log.info('Serving on %s', addrs)

    async def serve_forever(self):
        """Starts the server and waits until it is stopped."""
        if self._server is None:
            await self.start()
        async with self._server:
            await self._server.serve_forever()

    def close(self):
        """Stops the server."""
        if self._server:
            self._server.close()

    async def wait_closed(self):
        """Waits until server is fully closed."""
        if self._server:
            await self._server.wait_closed()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

import unittest
from astm import codec
from astm import constants
from astm.exceptions import NotAccepted
from astm.client import Client
from astm.tests.utils import DummyMixIn


class DummyClient(DummyMixIn, Client):

    def __init__(self, *args, **kwargs):
        super(DummyClient, self).__init__(*args, **kwargs)
        self.timeout = None

    def create_socket(self, family, type):
        pass

    def connect(self, address):
        pass


class emitter(object):

    def __init__(self, *args):
        self.outbox = list(args)
        self.pos = 0
        self.inbox = []

    def __iter__(self):
        return self

    def next(self):
        if self.pos >= len(self.outbox):
            raise StopIteration
        item = self.outbox[self.pos]
        self.pos += 1
        return item

    __next__ = next

    def send(self, value):
        self.inbox.append(value)
        return self.next()

    def put(self, record):
        self.outbox.append(record)


def simple_emitter():
    yield ['H']
    yield ['L']


class ClientTestCase(unittest.TestCase):

    def test_open_connection(self):
        client = DummyClient(simple_emitter)
        client.handle_connect()
        self.assertEqual(client.outbox[0], constants.ENQ)

    def test_fail_on_enq(self):
        client = DummyClient(emitter)
        self.assertRaises(NotAccepted, client.on_enq)

    def test_fail_on_eot(self):
        client = DummyClient(emitter)
        self.assertRaises(NotAccepted, client.on_eot)

    def test_fail_on_message(self):
        client = DummyClient(emitter)
        self.assertRaises(NotAccepted, client.on_message)

    def test_callback_on_sent_failure(self):
        def emitter():
            yield ['H']
            assert not (yield ['P'])
            yield ['L']
        client = DummyClient(emitter)
        client.handle_connect()
        client.on_ack()
        client.on_ack()
        client.on_nak()

    def test_emitter_may_send_new_record_after_nak_response(self):
        def emitter():
            yield ['H']
            assert (yield ['P'])
            ok = yield ['O']
            if not ok:
                yield ['R']
            yield ['L']
        client = DummyClient(emitter)
        client.handle_connect()
        client.on_ack()
        client.on_ack()
        client.on_ack()
        client.on_nak()
        self.assertEqual(client.outbox[-1][2:3], b'R')

    def test_empty_emitter(self):
        def emitter():
            if False:
                yield
        client = DummyClient(emitter)
        client.handle_connect()
        self.assertEqual(client.outbox[-1], constants.ENQ)
        client.on_ack()
        self.assertEqual(client.outbox[-2], constants.EOT)
        self.assertEqual(client.outbox[-1], None)

    def test_early_yield(self):
        def emitter():
            yield ['P']
            if False:
                yield ['H']
                yield ['L']
        client = DummyClient(emitter)
        client.handle_connect()
        self.assertRaises(AssertionError, client.on_ack)

    def test_late_ack(self):
        def emitter():
            if False:
                yield ['H']
                yield ['L']
        client = DummyClient(emitter)
        client.handle_connect()
        self.assertEqual(client.outbox[-1], constants.ENQ)
        client.on_ack()
        self.assertEqual(client.outbox[-2], constants.EOT)
        self.assertEqual(client.outbox[-1], None)
        client.on_ack()
        self.assertEqual(client.outbox[-1], None)

    def test_dummy_usage(self):
        def emitter():
            yield ['H']
            ok = yield ['P']
            assert ok
            ok = yield ['O']
            assert ok
            yield ['L']
        client = DummyClient(emitter)
        client.handle_connect()
        self.assertEqual(client.outbox[-1], constants.ENQ)
        client.on_ack()
        self.assertEqual(client.outbox[-1][1:3], b'1H')
        client.on_ack()
        self.assertEqual(client.outbox[-1][1:3], b'2P')
        client.on_ack()
        self.assertEqual(client.outbox[-1][1:3], b'3O')
        client.on_ack()
        self.assertEqual(client.outbox[-1][1:3], b'4L')
        client.on_ack()
        self.assertEqual(client.outbox[-2], constants.EOT)
        self.assertEqual(client.outbox[-1], constants.ENQ)
        client.on_ack()
        self.assertEqual(client.outbox[-2], constants.EOT)
        self.assertEqual(client.outbox[-1], None)

    def test_reject_header(self):
        def emitter():
            assert (yield ['H'])
            yield ['P']
            yield ['O']
            yield ['L']
        client = DummyClient(emitter)
        client.handle_connect()
        client.on_ack()
        self.assertRaises(AssertionError, client.on_nak)

    def test_nak_callback(self):
        def emitter():
            yield ['H']
            assert not (yield ['P'])
            yield ['L']
        client = DummyClient(emitter)
        client.handle_connect()
        client.on_ack()
        client.on_ack()
        client.on_nak()
        client.on_ack()

    def test_emit_after_nak(self):
        def emitter():
            yield ['H']
            assert not (yield ['P'])
            yield ['O']
            yield ['L']
        client = DummyClient(emitter)
        client.handle_connect()
        client.on_ack()
        client.on_ack()
        client.on_nak()
        client.on_ack()

    def test_terminate_on_exception_after_nake(self):
        def emitter():
            yield ['H']
            assert (yield ['P'])
            yield ['O']
            yield ['L']
        client = DummyClient(emitter)
        client.handle_connect()
        client.on_ack()
        client.on_ack()
        self.assertRaises(AssertionError, client.on_nak)
        self.assertEqual(client.outbox[-2], constants.EOT)
        self.assertEqual(client.outbox[-1], None)


    def test_messages_workflow(self):
        def emitter():
            yield ['H']
            yield ['C']
            yield ['P']
            yield ['O']
            yield ['O']
            yield ['P']
            yield ['C']
            yield ['O']
            yield ['O']
            yield ['C']
            yield ['R']
            yield ['C']
            yield ['R']
            yield ['R']
            yield ['L']
        client = DummyClient(emitter)
        client.handle_connect()
        client.on_ack()
        while client.outbox[-1] is not None:
            client.on_ack()

    def test_session_in_loop(self):
        def emitter():
            for i in range(2):
                yield ['H']
                yield ['P']
                yield ['O']
                yield ['L']
        client = DummyClient(emitter)
        client.handle_connect()
        client.on_ack()
        while client.outbox[-1] is not None:
            client.on_ack()
        self.assertEqual(list(client.outbox),
            [b'\x05',
             b'\x021H\r\x0389\r\n',
             b'\x022P\r\x0392\r\n',
             b'\x023O\r\x0392\r\n',
             b'\x024L\r\x0390\r\n',
             b'\x04',
             b'\x05',
             b'\x021H\r\x0389\r\n',
             b'\x022P\r\x0392\r\n',
             b'\x023O\r\x0392\r\n',
             b'\x024L\r\x0390\r\n',
             b'\x04',
             b'\x05',
             b'\x04',
             None])

    def test_reject_terminator(self):
        def emitter():
            assert (yield ['H'])
            assert (yield ['P'])
            assert (yield ['O'])
            assert (yield ['L'])
        client = DummyClient(emitter)
        client.handle_connect()
        client.on_ack()
        client.on_ack()
        client.on_ack()
        client.on_ack()
        self.assertEqual(client.outbox[-1][1:3], b'4L')
        self.assertRaises(AssertionError, client.on_nak)
        self.assertEqual(client.outbox[-2], constants.EOT)
        self.assertEqual(client.outbox[-1], None)

    def test_timeout_handler(self):
        def emitter():
            assert (yield ['H'])
            assert (yield ['P'])
            assert (yield ['O'])
            assert (yield ['L'])
        client = DummyClient(emitter)
        client.handle_connect()
        client.on_ack()
        client.on_timeout()
        self.assertEqual(client.outbox[-2], constants.EOT)
        self.assertEqual(client.outbox[-1], None)

    def test_chunked_response(self):
        def emitter():
            assert (yield ['H', 'foo', 'bar'])
            assert (yield ['L', 'bar', 'baz'])
        client = DummyClient(emitter, chunk_size=12)
        client.handle_connect()
        client.on_ack()
        self.assertTrue(codec.is_chunked_message(client.outbox[-1]))
        self.assertEqual(client.outbox[-1], b'\x021H|foo\x1750\r\n')
        client.on_ack()
        self.assertFalse(codec.is_chunked_message(client.outbox[-1]))
        self.assertEqual(client.outbox[-1], b'\x022|bar\r\x03F3\r\n')
        client.on_ack()
        self.assertTrue(codec.is_chunked_message(client.outbox[-1]))
        self.assertEqual(client.outbox[-1], b'\x023L|bar\x1747\r\n')
        client.on_ack()
        self.assertFalse(codec.is_chunked_message(client.outbox[-1]))
        self.assertEqual(client.outbox[-1], b'\x024|baz\r\x03FD\r\n')
        client.on_ack()
        self.assertEqual(client.outbox[-1], constants.ENQ)
        client.on_ack()
        self.assertEqual(client.outbox[-1], None)

    def test_bulk_mode(self):
        def emitter():
            assert (yield ['H', 'foo', 'bar'])
            assert (yield ['L', 'bar', 'baz'])
        client = DummyClient(emitter, chunk_size=12, bulk_mode=True)
        client.handle_connect()
        client.on_ack()
        self.assertTrue(codec.is_chunked_message(client.outbox[-1]))
        self.assertEqual(client.outbox[-1], b'\x021H|foo\x1750\r\n')
        client.on_ack()
        self.assertTrue(codec.is_chunked_message(client.outbox[-1]))
        self.assertEqual(client.outbox[-1], b'\x022|bar\r\x1707\r\n')
        client.on_ack()
        self.assertTrue(codec.is_chunked_message(client.outbox[-1]))
        self.assertEqual(client.outbox[-1], b'\x023L|bar\x1747\r\n')
        client.on_ack()
        self.assertFalse(codec.is_chunked_message(client.outbox[-1]))
        self.assertEqual(client.outbox[-1], b'\x024|baz\r\x03FD\r\n')
        client.on_ack()
        self.assertEqual(client.outbox[-1], constants.ENQ)
        client.on_ack()
        self.assertEqual(client.outbox[-1], None)



if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

#: ASTM specification base encoding.
ENCODING = 'latin-1'

#: Message start token.
STX = b'\x02'
#: Message end token.
ETX = b'\x03'
#: ASTM session termination token.
EOT = b'\x04'
#: ASTM session initialization token.
ENQ = b'\x05'
#: Command accepted token.
ACK = b'\x06'
#: Command rejected token.
NAK = b'\x15'
#: Message chunk end token.
ETB = b'\x17'
LF  = b'\x0A'
CR  = b'\x0D'
#: CR + LF shortcut.
CRLF = CR + LF

#: Message records delimiter.
RECORD_SEP    = b'\x0D' # \r #
#: Record fields delimiter.
FIELD_SEP     = b'\x7C' # |  #
#: Delimeter for repeated fields.
REPEAT_SEP    = b'\x5C' # \  #
#: Field components delimiter.
COMPONENT_SEP = b'\x5E' # ^  #
#: Date escape token.
ESCAPE_SEP    = b'\x26' # &  #



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

import os
import sys
import unittest

def suite():
    suite = unittest.TestSuite()
    for root, dirs, files in os.walk('.'):
        for file in files:
            if not (file.startswith('test_') and file.endswith('.py')):
                continue
            name = file.split('.')[0]
            modname = os.path.join(root, name).replace(os.path.sep, '.')
            modname = modname.lstrip('.')
            tests = unittest.defaultTestLoader.loadTestsFromName(modname)
            for test in tests:
                suite.addTests(test)
            sys.stdout.write('%s : %s tests%s'
                             % (modname, tests.countTestCases(), os.linesep))
            sys.stdout.flush()
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

from astm import __version__
from astm.mapping import (
    Component, ConstantField, ComponentField, DateField, DateTimeField,
    IntegerField, SetField, TextField
)
from astm.records import (
    HeaderRecord, PatientRecord, OrderRecord, ResultRecord, CommentRecord,
    TerminatorRecord
)

#: Information about sender.
#:
#: :param name: Name.
#: :type name: str
#:
#: :param version: Sender software version.
#: :type version: str
#:
Sender = Component.build(
    TextField(name='name', default='python-astm'),
    TextField(name='version', default=__version__)
)


#: Patient name structure.
#:
#: :param last: Last name. Length: 50.
#: :type last: str
#:
#: :param first: First name. Length: 50.
#: :type first: str
#:
PatientName = Component.build(
    TextField(name='last', length=50),
    TextField(name='first', length=50)
)


class Header(HeaderRecord):
    """ASTM header record.

    :param type: Record Type ID. Always ``H``.
    :type type: str

    :param delimeter: Delimiter Definition. Always ``\\^&``.
    :type delimeter: str

    :param message_id: Message Control ID. Not used.
    :type message_id: None

    :param password: Access Password. Not used.
    :type password: None

    :param sender: Information about sender. Optional.
    :type sender: :class:`Sender`

    :param address: Sender Street Address. Not used.
    :type address: None

    :param reserved: Reserved Field. Not used.
    :type reserved: None

    :param phone: Sender Telephone Number. Not used.
    :type phone: None

    :param chars: Sender Characteristics. Not used.
    :type chars: None

    :param receiver: Information about receiver. Not used.
    :type receiver: None

    :param comments: Comments. Not used.
    :type comments: None

    :param processing_id: Processing ID. Always ``P``.
    :type processing_id: str

    :param version: ASTM Version Number. Always ``E 1394-97``.
    :type version: str

    :param timestamp: Date and Time of Message
    :type timestamp: datetime.datetime
    """

    sender = ComponentField(Sender)
    processing_id = ConstantField(default='P')
    version = ConstantField(default='E 1394-97')


class CommonPatient(PatientRecord):
    """ASTM patient record.

    :param type: Record Type ID. Always ``P``.
    :type type: str

    :param seq: Sequence Number. Required.
    :type seq: int

    :param practice_id: Practice Assigned Patient ID. Required. Length: 12.
    :type practice_id: str

    :param laboratory_id: Laboratory Assigned Patient ID. Required. Length: 16.
    :type laboratory_id: str

    :param id: Patient ID. Not used.
    :type id: None

    :param name: Patient name.
    :type name: :class:`PatientName`

    :param maiden_name: Mother’s Maiden Name. Not used.
    :type maiden_name: None

    :param birthdate: Birthdate.
    :type birthdate: datetime.date

    :param sex: Patient Sex. One of: ``M`` (male), ``F`` (female),
                ``I`` (animal), ``None`` is unknown.
    :type sex: str

    :param race: Patient Race-Ethnic Origin. Not used.
    :type race: None

    :param address: Patient Address. Not used.
    :type address: None

    :param reserved: Reserved Field. Not used.
    :type reserved: None

    :param phone: Patient Telephone Number. Not used.
    :type phone: None

    :param physician_id: Attending Physician. Not used.
    :type physician_id: None

    :param special_1: Special Field #1. Not used.
    :type special_1: None

    :param special_2: Patient source. Possible values:
      - ``0``: internal patient;
      - ``1``: external patient.
    :type special_2: int

    :param height: Patient Height. Not used.
    :type height: None

    :param weight: Patient Weight. Not used.
    :type weight: None

    :param diagnosis: Patient’s Known Diagnosis. Not used.
    :type diagnosis: None

    :param medications: Patient’s Active Medications. Not used.
    :type medications: None

    :param diet: Patient’s Diet. Not used.
    :type diet: None

    :param practice_1: Practice Field No. 1. Not used.
    :type practice_1: None

    :param practice_2: Practice Field No. 2. Not used.
    :type practice_2: None

    :param admission_date: Admission/Discharge Dates. Not used.
    :type admission_date: None

    :param admission_status: Admission Status. Not used.
    :type admission_status: None

    :param location: Patient location. Length: 20.
    :type location: str

    :param diagnostic_code_nature: Nature of diagnostic code. Not used.
    :type diagnostic_code_nature: None

    :param diagnostic_code: Diagnostic code. Not used.
    :type diagnostic_code: None

    :param religion: Patient religion. Not used.
    :type religion: None

    :param martial_status: Martian status. Not used.
    :type martial_status: None

    :param isolation_status: Isolation status. Not used.
    :type isolation_status: None

    :param language: Language. Not used.
    :type language: None

    :param hospital_service: Hospital service. Not used.
    :type hospital_service: None

    :param hospital_institution: Hospital institution. Not used.
    :type hospital_institution: None

    :param dosage_category: Dosage category. Not used.
    :type dosage_category: None
    """
    birthdate = DateField()
    laboratory_id = TextField(required=True, length=16)
    location = TextField(length=20)
    name = ComponentField(PatientName)
    practice_id = TextField(required=True, length=12)
    sex = SetField(values=('M', 'F', None, 'I'))
    special_2 = SetField(values=(0, 1), field=IntegerField())


class CommonOrder(OrderRecord):
    biomaterial = TextField(length=20)
    laboratory_field_2 = TextField(length=12)
    priority = SetField(default='S', values=('S', 'R'))
    sample_id = TextField(required=True, length=12)
    user_field_1 = TextField(length=20)
    user_field_2 = TextField(length=1024)


class CommonResult(ResultRecord):
    completed_at = DateTimeField(required=True)
    value = TextField(required=True, length=20)


class CommonComment(CommentRecord):
    ctype = ConstantField(default='G')


class Terminator(TerminatorRecord):
    """ASTM terminator record.

    :param type: Record Type ID. Always ``L``.
    :type type: str

    :param seq: Sequential number. Always ``1``.
    :type seq: int

    :param code: Termination code. Always ``N``.
    :type code: str
    """



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

import os
import sys
import unittest
from astm.exceptions import NotAccepted, InvalidState
from astm.server import RequestHandler, BaseRecordsDispatcher
from astm import codec, constants, records
from astm.tests.utils import DummyMixIn, track_call


def null_dispatcher(*args, **kwargs):
    pass


class DummyRequestHandler(DummyMixIn, RequestHandler):
    dummy_dispatcher_called_time = 0

    def __init__(self, dispatcher=null_dispatcher):
        RequestHandler.__init__(self, None, dispatcher)

    def default_handler(self, data):
        return constants.NAK

    def dispatch(self, data):
        self.dummy_dispatcher_called_time += 1
        return super(DummyRequestHandler, self).dispatch(data)

    def recv(self, size):
        return codec.encode([records.HeaderRecord().to_astm()])[0]


class RequestHandlerTestCase(unittest.TestCase):

    def setUp(self):
        self.req = DummyRequestHandler()
        self.req.dispatcher = track_call(self.req.dispatcher)
        self.stderr = sys.stderr
        sys.stderr = open(os.devnull, 'w')

    def tearDown(self):
        sys.stderr.close()
        sys.stderr = self.stderr

    def test_allow_enq_only_for_init_state(self):
        self.assertEqual(self.req.on_enq(), constants.ACK)
        self.assertEqual(self.req.on_enq(), constants.NAK)

    def test_allow_eot_only_for_transfer_state(self):
        self.assertRaises(InvalidState, self.req.on_eot)
        self.req.on_enq()
        self.req.on_eot()

    def test_fail_on_enq(self):
        self.assertRaises(NotAccepted, self.req.on_ack)

    def test_fail_on_eot(self):
        self.assertRaises(NotAccepted, self.req.on_nak)

    def test_reject_message_on_invalid_state(self):
        self.assertEqual(self.req.on_message(), constants.NAK)

    def test_reject_message_on_handle_error(self):
        self.req.on_enq()
        self.assertEqual(self.req.on_message(), constants.NAK)
        self.assertFalse(self.req.dispatcher.was_called)

    def test_reject_message_on_dispatch_error(self):
        def dispatcher(message):
            codec.decode_message(message, 'latin-1')
        req = DummyRequestHandler(track_call(dispatcher))
        req.on_enq()
        req._last_recv_data = '|foo'.encode()
        self.assertEqual(req.on_message(), constants.NAK)
        self.assertTrue(req.dispatcher.was_called)

    def test_accept_message(self):
        self.req.on_enq()
        self.req._last_recv_data = codec.encode([records.HeaderRecord()
                                                        .to_astm()])[0]
        self.assertEqual(self.req.on_message(), constants.ACK)
        self.assertTrue(self.req.dispatcher.was_called)
        self.assertFalse(self.req._chunks)

    def test_handle_chunked_transfer(self):
        self.req.on_enq()
        self.req._chunks = [b'']
        self.req._last_recv_data = codec.encode([records.HeaderRecord()
                                                 .to_astm()])[0]
        self.assertEqual(self.req.on_message(), constants.ACK)
        self.assertTrue(self.req.dispatcher.was_called)
        self.assertFalse(self.req._chunks)

    def test_cleanup_input_buffer_on_message_reject(self):
        self.req.handle_read()
        self.assertEqual(self.req.dummy_dispatcher_called_time, 1)
        self.assertEqual(self.req.outbox[-1], constants.NAK)
        self.assertEqual(self.req._input_buffer, '')

    def test_close_on_timeout(self):
        self.req.close = track_call(self.req.close)
        self.req.on_timeout()
        self.assertTrue(self.req.close.was_called)


class RecordsDispatcherTestCase(unittest.TestCase):

    def setUp(self):
        d = BaseRecordsDispatcher()
        for key, value in d.dispatch.items():
            d.dispatch[key] = track_call(value)
        d.on_unknown = track_call(d.on_unknown)
        self.dispatcher = d

    def test_dispatch_header(self):
        message = codec.encode_message(1, ['H'], 'ascii')
        self.dispatcher(message)
        self.assertTrue(self.dispatcher.dispatch['H'].was_called)

    def test_dispatch_comment(self):
        message = codec.encode_message(1, ['C'], 'ascii')
        self.dispatcher(message)
        self.assertTrue(self.dispatcher.dispatch['C'].was_called)

    def test_dispatch_patient(self):
        message = codec.encode_message(1, ['P'], 'ascii')
        self.dispatcher(message)
        self.assertTrue(self.dispatcher.dispatch['P'].was_called)

    def test_dispatch_order(self):
        message = codec.encode_message(1, ['O'], 'ascii')
        self.dispatcher(message)
        self.assertTrue(self.dispatcher.dispatch['O'].was_called)

    def test_dispatch_result(self):
        message = codec.encode_message(1, ['R'], 'ascii')
        self.dispatcher(message)
        self.assertTrue(self.dispatcher.dispatch['R'].was_called)

    def test_dispatch_scientific(self):
        message = codec.encode_message(1, ['S'], 'ascii')
        self.dispatcher(message)
        self.assertTrue(self.dispatcher.dispatch['S'].was_called)

    def test_dispatch_manufacturer_info(self):
        message = codec.encode_message(1, ['M'], 'ascii')
        self.dispatcher(message)
        self.assertTrue(self.dispatcher.dispatch['M'].was_called)

    def test_dispatch_terminator(self):
        message = codec.encode_message(1, ['L'], 'ascii')
        self.dispatcher(message)
        self.assertTrue(self.dispatcher.dispatch['L'].was_called)

    def test_wrap_before_dispatch(self):
        class Thing(object):
            def __init__(self, *args):
                self.args = args
        def handler(record):
            assert isinstance(record, Thing)
        message = codec.encode_message(1, ['H'], 'ascii')
        self.dispatcher.wrappers['H'] = Thing
        self.dispatcher.dispatch['H'] = track_call(handler)
        self.dispatcher(message)
        self.assertTrue(self.dispatcher.dispatch['H'].was_called)

    def test_provide_default_handler_for_unknown_message_type(self):
        message = codec.encode_message(1, ['FOO'], 'ascii')
        self.dispatcher(message)
        self.assertTrue(self.dispatcher.on_unknown.was_called)



if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

import unittest
from astm import constants
from astm.tests.utils import DummyMixIn, track_call
from astm import protocol

class DummyTimer(object):
    def __init__(self, timeout, callback):
        self.timeout = timeout
        self.callback = callback
        self._alive = False

    def is_alive(self):
        return self._alive

    def start(self):
        self._alive = True

    def cancel(self):
        self._alive = False


class DummyProto(DummyMixIn, protocol.ASTMProtocol):
    _timer_cls = DummyTimer


class DispatcherTestCase(unittest.TestCase):

    def test_found_terminator(self):
        obj = DummyProto()
        obj.dispatch = track_call(obj.dispatch)
        obj.found_terminator()
        self.assertTrue(not obj.dispatch.was_called)

        obj.inbox.append(constants.ENQ)
        obj.found_terminator()
        self.assertTrue(obj.dispatch.was_called)

    def test_found_terminator_skip_empty(self):
        obj = DummyProto()
        obj.dispatch = track_call(obj.dispatch)
        obj.inbox.append('')
        obj.inbox.append(None)
        obj.found_terminator()
        self.assertTrue(not obj.dispatch.was_called)

    def test_on_enq(self):
        obj = DummyProto()
        obj.on_enq = track_call(obj.on_enq)
        obj.dispatch(constants.ENQ)
        self.assertTrue(obj.on_enq.was_called)

    def test_on_ack(self):
        obj = DummyProto()
        obj.on_ack = track_call(obj.on_ack)
        obj.dispatch(constants.ACK)
        self.assertTrue(obj.on_ack.was_called)

    def test_on_nak(self):
        obj = DummyProto()
        obj.on_nak = track_call(obj.on_nak)
        obj.dispatch(constants.NAK)
        self.assertTrue(obj.on_nak.was_called)

    def test_on_eot(self):
        obj = DummyProto()
        obj.on_eot = track_call(obj.on_eot)
        obj.dispatch(constants.EOT)
        self.assertTrue(obj.on_eot.was_called)

    def test_on_message(self):
        obj = DummyProto()
        obj.on_message = track_call(obj.on_message)
        obj.dispatch(constants.STX)
        self.assertTrue(obj.on_message.was_called)

    def test_default_hanlder(self):
        obj = DummyProto()
        obj.default_handler = track_call(obj.default_handler)
        self.assertRaises(ValueError, obj.dispatch, b'foo')
        self.assertTrue(obj.default_handler.was_called)

    def test_push_event_response(self):
        obj = DummyProto()
        obj.on_message = lambda: '42'
        obj.dispatch(constants.STX)
        self.assertEqual(obj.outbox.pop(), '42')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

class BaseASTMError(Exception):
    """Base ASTM error."""


class InvalidState(BaseASTMError):
    """Should be raised in case of invalid ASTM handler state."""


class NotAccepted(BaseException):
    """Received data is not acceptable."""


class Rejected(BaseASTMError):
    """Should be raised after unsuccessful attempts to send data
    (receiver sends with <NAK> reply)."""





# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

from . import client
from . import server



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#


class DummyMixIn(object):
    _input_buffer = ''
    addr = ('localhost', 15200)

    def flush(self):
        pass

    def close(self):
        pass


class CallLogger(object):

    def __init__(self, func):
        self.func = func
        self.was_called = False

    def __call__(self, *args, **kwargs):
        self.was_called = True
        return self.func(*args, **kwargs)


def track_call(func):
    return CallLogger(func)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

import datetime
import decimal
import inspect
import time
import warnings
from operator import itemgetter
from itertools import islice
from itertools import zip_longest
import logging
from functools import partial
from collections.abc import MutableMapping, Iterable

log = logging.getLogger(__name__)


def make_string(value):
    if isinstance(value, str):
        return value
    elif isinstance(value, bytes):
        return str(value, 'utf-8')
    else:
        return str(value)


class Field(object):
    """Base mapping field class."""
    def __init__(self, name=None, default=None, required=False, length=None):
        self.name = name
        self.default = default
        self.required = required
        self.length = length

    def __get__(self, instance, owner):
        if instance is None:
            return self
        value = instance._data.get(self.name)
        if value is not None:
            value = self._get_value(value)
        elif self.default is not None:
            default = self.default
            if hasattr(default, '__call__'):
                default = default()
            value = default
        return value

    def __set__(self, instance, value):
        if value is not None:
            value = self._set_value(value)
        instance._data[self.name] = value

    def _get_value(self, value):
        return value

    def _set_value(self, value):
        value = make_string(value)
        if self.length is not None and len(value) > self.length:
            raise ValueError('Field %r value is too long (max %d, got %d)'
                            '' % (self.name, self.length, len(value)))
        return value


class MetaMapping(type):

    def __new__(mcs, name, bases, d):
        fields = []
        names = []
        def merge_fields(items):
            for n, field in items:
                if field.name is None:
                    field.name = n
                if n not in names:
                    fields.append((n, field))
                    names.append(n)
                else:
                    fields[names.index(n)] = (n, field)
        for base in bases:
            if hasattr(base, '_fields'):
                merge_fields(base._fields)
        merge_fields([(k, v) for k, v in d.items() if isinstance(v, Field)])
        if '_fields' not in d:
            d['_fields'] = fields
        else:
            merge_fields(d['_fields'])
            d['_fields'] = fields
        return super(MetaMapping, mcs).__new__(mcs, name, bases, d)


class Mapping(metaclass=MetaMapping):

    def __init__(self, *args, **kwargs):
        fieldnames = map(itemgetter(0), self._fields)
        values = dict(zip_longest(fieldnames, args))
        values.update(kwargs)
        self._data = {}
        for attrname, field in self._fields:
            attrval = values.pop(attrname, None)
            if attrval is None:
                setattr(self, attrname, getattr(self, attrname))
            else:
                setattr(self, attrname, attrval)
        if values:
            raise ValueError('Unexpected kwargs found: %r' % values)

    @classmethod
    def build(cls, *a):
        fields = []
        newcls = type('Generic' + cls.__name__, (cls,), {})
        for field in a:
            if field.name is None:
                raise ValueError('Name is required for ordered fields.')
            setattr(newcls, field.name, field)
            fields.append((field.name, field))
        newcls._fields = fields
        return newcls

    def __getitem__(self, key):
        return self.values()[key]

    def __setitem__(self, key, value):
        setattr(self, self._fields[key][0], value)

    def __delitem__(self, key):
        self._data[self._fields[key][0]] = None

    def __iter__(self):
        return iter(self.values())

    def __contains__(self, item):
        return item in self.values()

    def __len__(self):
        return len(self._data)

    def __eq__(self, other):
        if len(self) != len(other):
            return False
        for key, value in zip(self.keys(), other):
            if getattr(self, key) != value:
                return False
        return True

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__,
                           ', '.join('%s=%r' % (key, value)
                                     for key, value in self.items()))

    def keys(self):
        return [key for key, field in self._fields]

    def values(self):
        return [getattr(self, key) for key in self.keys()]

    def items(self):
        return [(key, getattr(self, key)) for key, field in self._fields]

    def to_astm(self):
        def values(obj):
            for key, field in obj._fields:
                value = obj._data[key]
                if isinstance(value, Mapping):
                    yield list(values(value))
                elif isinstance(value, list):
                    stack = []
                    for item in value:
                        if isinstance(item, Mapping):
                            stack.append(list(values(item)))
                        else:
                            stack.append(item)
                    yield stack
                elif value is None and field.required:
                    raise ValueError('Field %r value should not be None' % key)
                else:
                    yield value
        return list(values(self))


class Record(Mapping):
    """ASTM record mapping class."""


class Component(Mapping):
    """ASTM component mapping class."""


class TextField(Field):
    """Mapping field for string values."""
    def _set_value(self, value):
        if not isinstance(value, str):
            raise TypeError('String value expected, got %r' % value)
        return super(TextField, self)._set_value(value)


class ConstantField(Field):
    def __init__(self, name=None, default=None, field=None):
        field = field or Field()
        super(ConstantField, self).__init__(name, default, True, None)
        self.field = field
        self.required = True
        if self.default is None:
            raise ValueError('Constant value should be defined')

    def _get_value(self, value):
        return self.default

    def _set_value(self, value):
        value = self.field._get_value(value)
        if self.default != value:
            raise ValueError('Field changing not allowed: got %r, accepts %r'
                            '' % (value, self.default))
        return super(ConstantField, self)._set_value(value)


class IntegerField(Field):
    """Mapping field for integer values."""
    def _get_value(self, value):
        return int(value)

    def _set_value(self, value):
        if not isinstance(value, int):
            try:
                value = self._get_value(value)
            except Exception:
                raise TypeError('Integer value expected, got %r' % value)
        return super(IntegerField, self)._set_value(value)


class DecimalField(Field):
    """Mapping field for decimal values."""
    def _get_value(self, value):
        return decimal.Decimal(value)

    def _set_value(self, value):
        if not isinstance(value, (int, float, decimal.Decimal)):
            raise TypeError('Decimal value expected, got %r' % value)
        return super(DecimalField, self)._set_value(value)


class DateField(Field):
    """Mapping field for storing date/time values."""
    format = '%Y%m%d'
    def _get_value(self, value):
        return datetime.datetime.strptime(value, self.format)

    def _set_value(self, value):
        if isinstance(value, str):
            value = self._get_value(value)
        if not isinstance(value, (datetime.datetime, datetime.date)):
            raise TypeError('Datetime value expected, got %r' % value)
        return value.strftime(self.format)


class TimeField(Field):
    """Mapping field for storing times."""
    format = '%H%M%S'
    def _get_value(self, value):
        if isinstance(value, str):
            try:
                value = value.split('.', 1)[0] # strip out microseconds
                value = datetime.time(*time.strptime(value, self.format)[3:6])
            except ValueError:
                raise ValueError('Value %r does not match format %s'
                                 '' % (value, self.format))
        return value

    def _set_value(self, value):
        if isinstance(value, str):
            value = self._get_value(value)
        if not isinstance(value, (datetime.datetime, datetime.time)):
            raise TypeError('Datetime value expected, got %r' % value)
        if isinstance(value, datetime.datetime):
            value = value.time()
        return value.replace(microsecond=0).strftime(self.format)


class DateTimeField(Field):
    """Mapping field for storing date/time values."""
    format = '%Y%m%d%H%M%S'
    def _get_value(self, value):
        return datetime.datetime.strptime(value, self.format)

    def _set_value(self, value):
        if isinstance(value, str):
            value = self._get_value(value)
        if not isinstance(value, (datetime.datetime, datetime.date)):
            raise TypeError('Datetime value expected, got %r' % value)
        return value.strftime(self.format)


class SetField(Field):
    """Mapping field for predefined set of values."""
    def __init__(self, name=None, default=None,
                 required=False, length=None,
                 values=None, field=None):
        field = field or Field()
        super(SetField, self).__init__(name, default, required, length)
        self.field = field
        self.values = values and set(values) or set([])

    def _get_value(self, value):
        return self.field._get_value(value)

    def _set_value(self, value):
        value = self.field._get_value(value)
        if value not in self.values:
            raise ValueError('Unexpectable value %r' % value)
        return self.field._set_value(value)


class ComponentField(Field):
    """Mapping field for storing record component."""
    def __init__(self, mapping, name=None, default=None):
        self.mapping = mapping
        default = default or mapping()
        super(ComponentField, self).__init__(name, default)


    def _get_value(self, value):
        if isinstance(value, dict):
            return self.mapping(**value)
        elif isinstance(value, self.mapping):
            return value
        else:
            return self.mapping(*value)

    def _set_value(self, value):
        if isinstance(value, dict):
            return self.mapping(**value)
        elif isinstance(value, self.mapping):
            return value
        if isinstance(value, str):
            value = [value]
        return self.mapping(*value)


class RepeatedComponentField(Field):
    """Mapping field for storing list of record components."""
    def __init__(self, field, name=None, default=None):
        if isinstance(field, ComponentField):
            self.field = field
        else:
            assert isinstance(field, type) and issubclass(field, Mapping)
            self.field = ComponentField(field)
        default = default or []
        super(RepeatedComponentField, self).__init__(name, default)

    class Proxy(list):
        def __init__(self, seq, field):
            super(RepeatedComponentField.Proxy, self).__init__(field._get_value(i) for i in seq)
            self.field = field

        def __setitem__(self, index, value):
            if isinstance(index, slice):
                super(RepeatedComponentField.Proxy, self).__setitem__(index, [self.field._set_value(v) for v in value])
            else:
                super(RepeatedComponentField.Proxy, self).__setitem__(index, self.field._set_value(value))

        def append(self, item):
            super(RepeatedComponentField.Proxy, self).append(self.field._set_value(item))

        def extend(self, other):
            super(RepeatedComponentField.Proxy, self).extend(self.field._set_value(i) for i in other)

    def _get_value(self, value):
        return self.Proxy(value, self.field)

    def _set_value(self, value):
        return [self.field._set_value(item) for item in value]


class NotUsedField(Field):
    def __init__(self, name=None):
        super(NotUsedField, self).__init__(name)

    def _get_value(self, value):
        return None

    def _set_value(self, value):
        warnings.warn('Field %r is not used, any assignments are omitted'
                      '' % self.name, UserWarning)
        return None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

"""

``astm.omnilab.server`` - LabOnline server implementation
----------------------------------------------------------

"""

from astm.server import BaseRecordsDispatcher
from astm.mapping import (
    Component, ConstantField, ComponentField, DateTimeField, IntegerField,
    SetField, TextField, NotUsedField, DateField
)
from .common import (
    Header, Terminator, CommonPatient as Patient,
    CommonOrder,
    CommonResult,
    CommonComment,
    Sender
)


__all__ = ['RecordsDispatcher',
           'Header', 'Patient', 'Order', 'Result', 'Terminator',
           'CommentData', 'CompletionDate', 'Instrument', 'Operator',
           'Sender', 'Test']

#: Instrument (analyser) information structure.
#:
#: :param _: Reserved. Not used.
#: :type _: None
#:
#: :param rack: Rack number. Length: 5.
#: :type rack: str
#:
#: :param position: Position number. Length: 3.
#: :type position: str
#:
Instrument = Component.build(
    TextField(name='_'),
    TextField(name='rack', length=5),
    TextField(name='position', length=3),
)


#: Test :class:`~astm.mapping.Component` also known as Universal Test ID.
#:
#: :param _: Reserved. Not used.
#: :type _: None
#:
#: :param __: Reserved. Not used.
#: :type __: None
#:
#: :param ___: Reserved. Not used.
#: :type ___: None
#:
#: :param assay_code: Assay code. Required. Length: 20.
#: :type assay_code: str
#:
#: :param assay_name: Assay name. Length: 8.
#: :type assay_name: str
#:
#: :param dilution: Dilution. Length: 10.
#: :type dilution: str
#:
#: :param status: Assay status. Length: 1.
#: :type status: str
#:
#: :param reagent_lot: Reagent lot. Length: 15.
#: :type reagent_lot: str
#:
#: :param reagent_number: Reagent serial number. Length: 5.
#: :type reagent_number: str
#:
#: :param control_lot: Control lot number. Length: 25.
#: :type control_lot: str
#:
#: :param type: Result type value. One of: ``CE``, ``TX``.
#: :type type: str
#:
Test = Component.build(
    NotUsedField(name='_'),
    NotUsedField(name='__'),
    NotUsedField(name='___'),
    TextField(name='assay_code', required=True, length=20),
    TextField(name='assay_name', length=8),
    TextField(name='dilution', length=10),
    TextField(name='status', length=1),
    TextField(name='reagent_lot', length=15),
    TextField(name='reagent_number', length=5),
    TextField(name='control_lot', length=25),
    SetField(name='type', values=('CE', 'TX'))
)


#: Information about operator that validated results.
#:
#: :param code_on_labonline: Operator code on LabOnline. Length: 12.
#: :type code_on_labonline: str
#:
#: :param code_on_analyzer: Operator code on analyser. Length: 20.
#: :type code_on_analyzer: str
#:
Operator = Component.build(
    TextField(name='code_on_labonline', length=12),
    TextField(name='code_on_analyzer', length=20),
)


#: Completion date time information.
#:
#: :param labonline: Completion date time on LabOnline.
#: :type labonline: datetime.datetime
#:
#: :param analyzer: Completion date time on analyser.
#: :type analyzer: datetime.datetime
#:
CompletionDate = Component.build(
    DateTimeField(name='labonline'),
    DateTimeField(name='analyzer'),
)

#: Instrument (analyser) information structure.
#:
#: :param _: Reserved. Not used.
#: :type _: None
#:
#: :param rack: Rack number. Length: 5.
#: :type rack: str
#:
#: :param position: Position number. Length: 3.
#: :type position: str
#:
Instrument = Component.build(
    NotUsedField(name='_'),
    TextField(name='rack', length=5),
    TextField(name='position', length=3),
)

#: Comment control text structure.
#:
CommentData = Component.build(
    SetField(name='code', values=('PC', 'RC', 'SC', 'TC',
        'CK', 'SE', 'CL', 'TA', 'SS', 'HQ', 'AL', 'PT')),
    TextField(name='value'),
    TextField(name='field_1'),
    TextField(name='field_2'),
    TextField(name='field_3'),
    TextField(name='field_4'),
    TextField(name='field_5'),
)


class Order(CommonOrder):
    """ASTM order record.

    :param type: Record Type ID. Always ``O``.
    :type type: str

    :param seq: Sequence Number. Required.
    :type seq: int

    :param sample_id: Sample ID number. Required. Length: 12.
    :type sample_id: str

    :param instrument: Instrument specimen ID.
    :type instrument: :class:`Instrument`

    :param test: Test information structure (aka Universal Test ID).
    :type test: :class:`Test`

    :param priority: Priority flag. Required. Possible values:
                     - ``S``: stat; -``R``: routine.
    :type priority: str

    :param created_at: Ordered date and time. Required.
    :type created_at: datetime.datetime

    :param sampled_at: Specimen collection date and time.
    :type sampled_at: datetime.datetime

    :param collected_at: Collection end time. Not used.
    :type collected_at: None

    :param volume: Collection volume. Not used.
    :type volume: None

    :param collector: Collector ID. Not used.
    :type collector: None

    :param action_code: Action code. Required. Possible values:
                        - :const:`None`: normal order result;
                        - ``Q``: quality control;
    :type action_code: str

    :param danger_code: Danger code. Not used.
    :type danger_code: None

    :param clinical_info: Revelant clinical info. Not used.
    :type clinical_info: None

    :param delivered_at: Date/time specimen received.
    :type delivered_at: None

    :param biomaterial: Sample material code. Length: 20.
    :type biomaterial: str

    :param physician: Ordering Physician. Not used.
    :type physician: None

    :param physician_phone: Physician's phone number. Not used.
    :type physician_phone: None

    :param user_field_1: An optional field, it will be send back unchanged to
                         the host along with the result. Length: 20.
    :type user_field_1: str

    :param user_field_2: An optional field, it will be send back unchanged to
                         the host along with the result. Length: 1024.
    :type user_field_2: str

    :param laboratory_field_1: Laboratory field #1. Not used.
    :type laboratory_field_1: None

    :param laboratory_field_2: Primary tube code. Length: 12.
    :type laboratory_field_2: str

    :param modified_at: Date and time of last result modification. Not used.
    :type modified_at: None

    :param instrument_charge: Instrument charge to computer system. Not used.
    :type instrument_charge: None

    :param instrument_section: Instrument section id. Not used.
    :type instrument_section: None

    :param report_type: Report type. Always ``F`` which means final order
                        request.
    :type report_type: str

    :param reserved: Reserved. Not used.
    :type reserved: None

    :param location_ward: Location ward of specimen collection. Not used.
    :type location_ward: None

    :param infection_flag: Nosocomial infection flag. Not used.
    :type infection_flag: None

    :param specimen_service: Specimen service. Not used.
    :type specimen_service: None

    :param laboratory: Production laboratory. Not used.
    :type laboratory: None
    """
    action_code = SetField(values=(None, 'Q'))
    instrument = ComponentField(Instrument)
    report_type = ConstantField(default='F')
    test = ComponentField(Test)


class Result(CommonResult):
    """ASTM patient record.

    :param type: Record Type ID. Always ``R``.
    :type type: str

    :param seq: Sequence Number. Required.
    :type seq: int

    :param test: Test information structure (aka Universal Test ID).
    :type test: :class:`Test`

    :param value: Measurement value. Numeric, coded or free text value
                  depending on result type. Required. Length: 1024.
    :type value: None

    :param units: Units. Length: 20.
    :type units: str

    :param references: Normal reference value interval.
    :type references: str

    :param abnormal_flag: Result abnormal flag. Possible values:
                          - ``0``: normal result;
                          - ``1``: result out of normal values;
                          - ``2``: result out of attention values;
                          - ``3``: result out of panic values;
                          +10 Delta-check;
                          +1000 Device alarm.
                          Length: 4.
    :type abnormal_flag: str

    :param abnormality_nature: Nature of abnormality testing. Possible values:
                               - ``N``: normal value;
                               - ``L``: below low normal range;
                               - ``H``: above high normal range;
                               - ``LL``: below low critical range;
                               - ``HH``: above high critical range.
    :type abnormality_nature: str

    :param status: Result status. ``F`` indicates a final result;
                   ``R`` indicating rerun. Length: 1.
    :type status: str

    :param normatives_changed_at: Date of changes in instrument normative
                                  values or units. Not used.
    :type normatives_changed_at: None

    :param operator: Operator ID.
    :type operator: :class:`Operator`

    :param started_at: When works on test was started on.
    :type started_at: datetime.datetime

    :param completed_at: When works on test was done.
    :type completed_at: datetime.datetime

    :param instrument: Instrument ID. Required.
    :type instrument: :class:`Instrument`
    """
    abnormal_flag = SetField(
        field=IntegerField(),
        length=4,
        values=(0, 1, 2, 3,
                10, 11, 12, 13,
                1000, 1001, 1002, 1003,
                1010, 1011, 1012, 1013))
    abnormality_nature = SetField(values=('N', 'L', 'H', 'LL', 'HH'))
    completed_at = ComponentField(CompletionDate)
    created_at = DateField()
    instrument = TextField(length=16)
    operator = ComponentField(Operator)
    references = TextField()
    sampled_at = DateField()
    started_at = DateTimeField(required=True)
    status = SetField(values=('F', 'R'))
    test = ComponentField(Test)
    units = TextField(length=20)


class Comment(CommonComment):
    """ASTM patient record.

    :param type: Record Type ID. Always ``C``.
    :type type: str

    :param seq: Sequence Number. Required.
    :type seq: int

    :param source: Comment source. Always ``I``.
    :type source: str

    :param data: Measurement value. Numeric, coded or free text value
                  depending on result type. Required. Length: 1024.
    :type data: :class:`CommentData`

    :param ctype: Comment type. Always ``G``.
    :type ctype: str
    """
    source = ConstantField(default='I')
    data = ComponentField(CommentData)


class RecordsDispatcher(BaseRecordsDispatcher):
    """Omnilab specific records dispatcher. Automatically wraps records by
    related mappings."""
    def __init__(self, *args, **kwargs):
        super(RecordsDispatcher, self).__init__(*args, **kwargs)
        self.wrappers = {
            'H': Header,
            'P': Patient,
            'O': Order,
            'R': Result,
            'C': Comment,
            'L': Terminator
        }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#
import asyncio
import logging
from .codec import encode
from .constants import ACK, ENQ, EOT, NAK, ENCODING
from .exceptions import NotAccepted
from .mapping import Record

log = logging.getLogger(__name__)

__all__ = ['Client']


class Client:
    """
    Asynchronous ASTM client.

    :param host: Server IP address or hostname.
    :type host: str

    :param port: Server port number.
    :type port: int

    :param encoding: Data encoding.
    :type encoding: str

    :param timeout: Time to wait for response from server in seconds.
    :type timeout: int
    """
    encoding = ENCODING

    def __init__(self, host='localhost', port=15200,
                 encoding=None, timeout=10):
        self.host = host
        self.port = port
        self.encoding = encoding or self.encoding
        self.timeout = timeout
        self._reader = None
        self._writer = None

    async def connect(self):
        """
        Connects to the server.
        """
        self._reader, self._writer = await asyncio.open_connection(
            self.host, self.port)
        peername = self._writer.get_extra_info('peername')
        log.info('Connection established to %s', peername)

    async def _read(self):
        try:
            return await asyncio.wait_for(self._reader.read(1), self.timeout)
        except asyncio.TimeoutError:
            log.error('Connection timed out.')
            self.close()
            await self.wait_closed()
            return None

    async def send(self, records, chunk_size=None):
        """
        Sends ASTM records to the server.

        :param records: An iterable of ASTM records. Each record should be a
                        list or a :class:`~astm.mapping.Record` object.
        :param chunk_size: The size of each data chunk in bytes. If `None`,
                           the data will not be chunked.
        :type chunk_size: int or None

        :return: `True` if all records were sent and acknowledged, `False`
                 otherwise.
        :rtype: bool
        """
        if not (self._reader and self._writer):
            await self.connect()

        self._writer.write(ENQ)
        await self._writer.drain()

        response = await self._read()
        if response != ACK:
            log.error('Server did not acknowledge session start.')
            self._writer.write(EOT)
            await self._writer.drain()
            return False

        messages = encode(records, encoding=self.encoding, chunk_size=chunk_size)

        for message in messages:
            self._writer.write(message)
            await self._writer.drain()
            response = await self._read()
            if response != ACK:
                log.error('Server did not acknowledge message: %r', message)
                self._writer.write(EOT)
                await self._writer.drain()
                return False
        
        self._writer.write(EOT)
        await self._writer.drain()
        log.info('Session finished successfully.')
        return True

    def close(self):
        """
        Closes the connection to the server.
        """
        if self._writer:
            self._writer.close()

    async def wait_closed(self):
        """
        Waits until the connection is fully closed.
        """
        if self._writer:
            await self._writer.wait_closed()
            self._reader = None
            self._writer = None
            log.info('Connection closed.')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

"""

``astm.omnilab.client`` - LabOnline client implementation
----------------------------------------------------------

"""

from astm.mapping import (
    Component, ConstantField, ComponentField, IntegerField, DateTimeField,
    RepeatedComponentField, SetField, TextField, NotUsedField
)
from .common import (
    Header, Terminator,
    CommonPatient,
    CommonOrder,
    CommonResult,
    CommonComment,
    Sender
)

__all__ = ['Header', 'Patient', 'Order', 'Result', 'Comment', 'Terminator',
           'CommentData', 'PatientAge', 'Sender', 'Test']

#: Patient age structure.
#:
#: :param value: Age value.
#: :type value: int
#:
#: :param unit: Age unit. One of: ``years``, ``months``, ``days``.
#: :type unit: str
#:
PatientAge = Component.build(
    IntegerField(name='value'),
    SetField(name='unit', values=('years', 'months', 'days'))
)

#: Test :class:`~astm.mapping.Component` also known as Universal Test ID.
#:
#: :param _: Reserved. Not used.
#: :type _: None
#:
#: :param __: Reserved. Not used.
#: :type __: None
#:
#: :param ___: Reserved. Not used.
#: :type ___: None
#:
#: :param assay_code: Assay code. Required. Length: 20.
#: :type assay_code: str
#:
#: :param assay_name: Assay name. Length: 8.
#: :type assay_name: str
#:
Test = Component.build(
    NotUsedField(name='_'),
    NotUsedField(name='__'),
    NotUsedField(name='___'),
    TextField(name='assay_code', required=True, length=20),
    TextField(name='assay_name', length=8),
)

#: Comment control data structure.
#:
CommentData = Component.build(
    SetField(name='code', values=('PC', 'RC', 'SC', 'TC')),
    TextField(name='value')
)


class Patient(CommonPatient):
    """ASTM patient record.

    :param type: Record Type ID. Always ``P``.
    :type type: str

    :param seq: Sequence Number. Required.
    :type seq: int

    :param practice_id: Practice Assigned Patient ID. Required. Length: 12.
    :type practice_id: str

    :param laboratory_id: Laboratory Assigned Patient ID. Required. Length: 16.
    :type laboratory_id: str

    :param id: Patient ID. Not used.
    :type id: None

    :param name: Patient name.
    :type name: :class:`PatientName`

    :param maiden_name: Mother’s Maiden Name. Not used.
    :type maiden_name: None

    :param birthdate: Birthdate.
    :type birthdate: datetime.date

    :param sex: Patient Sex. One of: ``M`` (male), ``F`` (female),
                ``I`` (animal), ``None`` is unknown.
    :type sex: str

    :param race: Patient Race-Ethnic Origin. Not used.
    :type race: None

    :param address: Patient Address. Not used.
    :type address: None

    :param reserved: Reserved Field. Not used.
    :type reserved: None

    :param phone: Patient Telephone Number. Not used.
    :type phone: None

    :param physician_id: Attending Physician. Not used.
    :type physician_id: None

    :param special_1: Special Field #1. Not used.
    :type special_1: None

    :param special_2: Patient source. Possible values:
      - ``0``: internal patient;
      - ``1``: external patient.
    :type special_2: int

    :param height: Patient Height. Not used.
    :type height: None

    :param weight: Patient Weight. Not used.
    :type weight: None

    :param diagnosis: Patient’s Known Diagnosis. Not used.
    :type diagnosis: None

    :param medications: Patient’s Active Medications. Not used.
    :type medications: None

    :param diet: Patient’s Diet. Not used.
    :type diet: None

    :param practice_1: Practice Field No. 1. Not used.
    :type practice_1: None

    :param practice_2: Practice Field No. 2. Not used.
    :type practice_2: None

    :param admission_date: Admission/Discharge Dates. Not used.
    :type admission_date: None

    :param admission_status: Admission Status. Not used.
    :type admission_status: None

    :param location: Patient location. Length: 20.
    :type location: str

    :param diagnostic_code_nature: Nature of diagnostic code. Not used.
    :type diagnostic_code_nature: None

    :param diagnostic_code: Diagnostic code. Not used.
    :type diagnostic_code: None

    :param religion: Patient religion. Not used.
    :type religion: None

    :param martial_status: Martian status. Not used.
    :type martial_status: None

    :param isolation_status: Isolation status. Not used.
    :type isolation_status: None

    :param language: Language. Not used.
    :type language: None

    :param hospital_service: Hospital service. Not used.
    :type hospital_service: None

    :param hospital_institution: Hospital institution. Not used.
    :type hospital_institution: None

    :param dosage_category: Dosage category. Not used.
    :type dosage_category: None
    """
    physician_id = TextField(length=35)
    special_1 = ComponentField(PatientAge)


class Order(CommonOrder):
    """ASTM order record.

    :param type: Record Type ID. Always ``O``.
    :type type: str

    :param seq: Sequence Number. Required.
    :type seq: int

    :param sample_id: Sample ID number. Required. Length: 12.
    :type sample_id: str

    :param instrument: Instrument specimen ID. Not used.
    :type instrument: None

    :param test: Test information structure (aka Universal Test ID).
    :type test: :class:`Test`

    :param priority: Priority flag. Required. Possible values:
                     - ``S``: stat; -``R``: routine.
    :type priority: str

    :param created_at: Ordered date and time. Required.
    :type created_at: datetime.datetime

    :param sampled_at: Specimen collection date and time.
    :type sampled_at: datetime.datetime

    :param collected_at: Collection end time. Not used.
    :type collected_at: None

    :param volume: Collection volume. Not used.
    :type volume: None

    :param collector: Collector ID. Not used.
    :type collector: None

    :param action_code: Action code. Required. Possible values:
                        - ``C``: cancel works for specified tests;
                        - ``A``: add tests to existed specimen;
                        - ``N``: create new order;
                        - ``R``: rerun tests for specified order;
    :type action_code: str

    :param danger_code: Danger code. Not used.
    :type danger_code: None

    :param clinical_info: Revelant clinical info. Not used.
    :type clinical_info: None

    :param delivered_at: Date/time specimen received.
    :type delivered_at: None

    :param biomaterial: Sample material code. Length: 20.
    :type biomaterial: str

    :param physician: Ordering Physician. Not used.
    :type physician: None

    :param physician_phone: Physician's phone number. Not used.
    :type physician_phone: None

    :param user_field_1: An optional field, it will be send back unchanged to
                         the host along with the result. Length: 20.
    :type user_field_1: str

    :param user_field_2: An optional field, it will be send back unchanged to
                         the host along with the result. Length: 1024.
    :type user_field_2: str

    :param laboratory_field_1: In multi-laboratory environment it will be used
                               to indicate which laboratory entering the order.
                               Length: 20.
    :type laboratory_field_1: str

    :param laboratory_field_2: Primary tube code. Length: 12.
    :type laboratory_field_2: str

    :param modified_at: Date and time of last result modification. Not used.
    :type modified_at: None

    :param instrument_charge: Instrument charge to computer system. Not used.
    :type instrument_charge: None

    :param instrument_section: Instrument section id. Not used.
    :type instrument_section: None

    :param report_type: Report type. Always ``O`` which means normal order
                        request.
    :type report_type: str

    :param reserved: Reserved. Not used.
    :type reserved: None

    :param location_ward: Location ward of specimen collection. Not used.
    :type location_ward: None

    :param infection_flag: Nosocomial infection flag. Not used.
    :type infection_flag: None

    :param specimen_service: Specimen service. Not used.
    :type specimen_service: None

    :param laboratory: Production laboratory: in multi-laboratory environment
                       indicates laboratory expected to process the order.
                       Length: 20.
    :type laboratory: str
    """
    action_code = SetField(default='N', values=('C', 'A', 'N', 'R'))
    created_at = DateTimeField(required=True)
    laboratory = TextField(length=20)
    laboratory_field_1 = TextField(length=20)
    report_type = ConstantField(default='O')
    sampled_at = DateTimeField()
    test = RepeatedComponentField(Test)


class Result(CommonResult):
    """ASTM patient record.

    :param type: Record Type ID. Always ``R``.
    :type type: str

    :param seq: Sequence Number. Required.
    :type seq: int

    :param test: Test information structure (aka Universal Test ID).
    :type test: :class:`Test`

    :param value: Measurement value. Numeric, coded or free text value
                  depending on result type. Required. Length: 1024.
    :type value: None

    :param units: Units. Not used.
    :type units: None

    :param references: Reference ranges. Not used.
    :type references: None

    :param abnormal_flag: Result abnormal flag. Not used.
    :type abnormal_flag: None

    :param abnormality_nature: Nature of abnormality testing. Not used.
    :type abnormality_nature: None

    :param status: Result status. Not used.
    :type status: None

    :param normatives_changed_at: Date of changes in instrument normative
                                  values or units. Not used.
    :type normatives_changed_at: None

    :param operator: Operator ID. Not used.
    :type operator: None

    :param started_at: When works on test was started on. Not used.
    :type started_at: None

    :param completed_at: When works on test was done. Required.
    :type completed_at: datetime.datetime

    :param instrument: Instrument ID. Not used.
    :type instrument: None
    """
    test = ComponentField(Test)


class Comment(CommonComment):
    """ASTM patient record.

    :param type: Record Type ID. Always ``C``.
    :type type: str

    :param seq: Sequence Number. Required.
    :type seq: int

    :param source: Comment source. Always ``L``.
    :type source: str

    :param data: Measurement value. Numeric, coded or free text value
                  depending on result type. Required. Length: 1024.
    :type data: :class:`CommentData`

    :param ctype: Comment type. Always ``G``.
    :type ctype: str
    """
    source = ConstantField(default='L')
    data = ComponentField(CommentData)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

__version_info__ = (1, 0, 0)
__version__ = '.'.join(map(str, __version_info__))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

from .version import __version__, __version_info__
from .exceptions import BaseASTMError, NotAccepted, InvalidState
from .codec import (
    decode, decode_message, decode_record,
    encode, encode_message, encode_record,
    make_checksum
)
from .mapping import Record, Component
from .records import (
    HeaderRecord, PatientRecord, OrderRecord,
    ResultRecord, CommentRecord, TerminatorRecord
)
from .protocol import ASTMProtocol
from .client import Client
from .server import RequestHandler, Server

import logging
log = logging.getLogger()

class NullHandler(logging.Handler):
    def emit(self, *args, **kwargs):
        pass

log.addHandler(NullHandler())



from astm import asynclib
import unittest
import select
import os
import socket
import threading
import time
import errno
from io import BytesIO
from unittest import skip

HOST = "127.0.0.1"
TESTFN = "@test"

class dummysocket:
    def __init__(self):
        self.closed = False

    def close(self):
        self.closed = True

    def fileno(self):
        return 42

class dummychannel:
    def __init__(self):
        self.socket = dummysocket()

    def close(self):
        self.socket.close()

class exitingdummy:
    def __init__(self):
        pass

    def handle_read_event(self):
        raise asynclib.ExitNow()

    handle_write_event = handle_read_event
    handle_close = handle_read_event
    handle_exception_event = handle_read_event

class crashingdummy:
    def __init__(self):
        self.error_handled = False

    def handle_read_event(self):
        raise Exception()

    handle_write_event = handle_read_event
    handle_close = handle_read_event
    handle_exception_event = handle_read_event

    def handle_error(self):
        self.error_handled = True

class dispatcherwithsend_noread(asynclib.Dispatcher):

    def __init__(self, sock=None, map=None):
        super(dispatcherwithsend_noread, self).__init__(sock, map)
        self.out_buffer = ''

    def initiate_send(self):
        num_sent = super(dispatcherwithsend_noread, self).send(self.out_buffer[:512])
        self.out_buffer = self.out_buffer[num_sent:]

    def handle_write(self):
        self.initiate_send()

    def writable(self):
        return (not self.connected) or len(self.out_buffer)

    def send(self, data):
        self.out_buffer += data.decode()
        self.initiate_send()

    def readable(self):
        return False

    def handle_connect(self):
        pass

# used when testing senders; just collects what it gets until newline is sent
def capture_server(evt, buf, serv):
    try:
        serv.listen(5)
        conn, addr = serv.accept()
    except socket.timeout:
        pass
    else:
        n = 200
        while n > 0:
            r, w, e = select.select([conn], [], [])
            if r:
                data = conn.recv(10)
                # keep everything except for the newline terminator
                buf.write(data.replace('\n', ''))
                if '\n' in data:
                    break
            n -= 1
            time.sleep(0.01)

        conn.close()
    finally:
        serv.close()
        evt.set()


class HelperFunctionTests(unittest.TestCase):
    def test_readwriteexc(self):
        # Check exception handling behavior of read, write and _exception

        # check that ExitNow exceptions in the object handler method
        # bubbles all the way up through asynclib read/write/_exception calls
        tr1 = exitingdummy()
        self.assertRaises(asynclib.ExitNow, asynclib.read, tr1)
        self.assertRaises(asynclib.ExitNow, asynclib.write, tr1)
        self.assertRaises(asynclib.ExitNow, asynclib.exception, tr1)

        # check that an exception other than ExitNow in the object handler
        # method causes the handle_error method to get called
        tr2 = crashingdummy()
        asynclib.read(tr2)
        self.assertEqual(tr2.error_handled, True)

        tr2 = crashingdummy()
        asynclib.write(tr2)
        self.assertEqual(tr2.error_handled, True)

        tr2 = crashingdummy()
        asynclib.exception(tr2)
        self.assertEqual(tr2.error_handled, True)

    # asynclib.readwrite uses constants in the select module that
    # are not present in Windows systems (see this thread:
    # http://mail.python.org/pipermail/python-list/2001-October/109973.html)
    # These constants should be present as long as poll is available

    if hasattr(select, 'poll'):
        def test_readwrite(self):
            # Check that correct methods are called by readwrite()

            attributes = ('read', 'expt', 'write', 'closed', 'error_handled')

            expected = (
                (select.POLLIN, 'read'),
                (select.POLLPRI, 'expt'),
                (select.POLLOUT, 'write'),
                (select.POLLERR, 'closed'),
                (select.POLLHUP, 'closed'),
                (select.POLLNVAL, 'closed'),
                )

            class testobj:
                def __init__(self):
                    self.read = False
                    self.write = False
                    self.closed = False
                    self.expt = False
                    self.error_handled = False

                def handle_read_event(self):
                    self.read = True

                def handle_write_event(self):
                    self.write = True

                def handle_close(self):
                    self.closed = True

                def handle_exception_event(self):
                    self.expt = True

                def handle_error(self):
                    self.error_handled = True

            for flag, expectedattr in expected:
                tobj = testobj()
                self.assertEqual(getattr(tobj, expectedattr), False)
                asynclib.readwrite(tobj, flag)

                # Only the attribute modified by the routine we expect to be
                # called should be True.
                for attr in attributes:
                    self.assertEqual(getattr(tobj, attr), attr==expectedattr)

                # check that ExitNow exceptions in the object handler method
                # bubbles all the way up through asynclib readwrite call
                tr1 = exitingdummy()
                self.assertRaises(asynclib.ExitNow, asynclib.readwrite, tr1, flag)

                # check that an exception other than ExitNow in the object handler
                # method causes the handle_error method to get called
                tr2 = crashingdummy()
                self.assertEqual(tr2.error_handled, False)
                asynclib.readwrite(tr2, flag)
                self.assertEqual(tr2.error_handled, True)

    def test_closeall(self):
        self.closeall_check(False)

    def test_closeall_default(self):
        self.closeall_check(True)

    def closeall_check(self, usedefault):
        # Check that close_all() closes everything in a given map

        l = []
        testmap = {}
        for i in range(10):
            c = dummychannel()
            l.append(c)
            self.assertEqual(c.socket.closed, False)
            testmap[i] = c

        if usedefault:
            socketmap = asynclib._SOCKET_MAP
            try:
                asynclib._SOCKET_MAP = testmap
                asynclib.close_all()
            finally:
                testmap, asynclib._SOCKET_MAP = asynclib._SOCKET_MAP, socketmap
        else:
            asynclib.close_all(testmap)

        self.assertEqual(len(testmap), 0)

        for c in l:
            self.assertEqual(c.socket.closed, True)


class DispatcherTests(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        asynclib.close_all()

    def test_basic(self):
        d = asynclib.Dispatcher()
        self.assertEqual(d.readable(), True)
        self.assertEqual(d.writable(), True)

    def test_repr(self):
        d = asynclib.Dispatcher()
        self.assertTrue(repr(d).endswith('Dispatcher at %#x>' % id(d)))

    def test_strerror(self):
        # refers to bug #8573
        err = asynclib._strerror(errno.EPERM)
        if hasattr(os, 'strerror'):
            self.assertEqual(err, os.strerror(errno.EPERM))
        err = asynclib._strerror(-1)
        self.assertTrue("unknown error" in err.lower())


class DispatcherWithSendTests(unittest.TestCase):
    usepoll = False

    def setUp(self):
        pass

    def tearDown(self):
        asynclib.close_all()

    @skip
    def test_send(self):
        self.evt = threading.Event()
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.settimeout(3)
        self.port = test_support.bind_port(self.sock)

        cap = StringIO()
        args = (self.evt, cap, self.sock)
        threading.Thread(target=capture_server, args=args).start()

        # wait a little longer for the server to initialize (it sometimes
        # refuses connections on slow machines without this wait)
        time.sleep(0.2)

        data = "Suppose there isn't a 16-ton weight?".encode()
        d = dispatcherwithsend_noread()
        d.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        d.connect((HOST, self.port))

        # give time for socket to connect
        time.sleep(0.1)

        d.send(data)
        d.send(data)
        d.send('\n')

        n = 1000
        while d.out_buffer and n > 0:
            asynclib.poll()
            n -= 1

        self.evt.wait()

        self.assertEqual(cap.getvalue(), data*2)


class DispatcherWithSendTests_UsePoll(DispatcherWithSendTests):
    usepoll = True


class CallLaterTests(unittest.TestCase):
    """Tests for CallLater class."""

    def setUp(self):
        # remove any unfired scheduled call left behind
        asynclib.close_all()

    def scheduler(self, timeout=0.01, count=100):
        while asynclib._SCHEDULED_TASKS and count > 0:
            asynclib.scheduler()
            count -= 1
            time.sleep(timeout)

    def test_interface(self):
        fun = lambda: 0
        self.assertRaises(AssertionError, asynclib.call_later, -1, fun)
        x = asynclib.call_later(3, fun)
        self.assertRaises(AssertionError, x.delay, -1)
        self.assertTrue(x.cancelled is False)
        x.cancel()
        self.assertTrue(x.cancelled is True)
        self.assertRaises(AssertionError, x.call)
        self.assertRaises(AssertionError, x.reset)
        self.assertRaises(AssertionError, x.delay, 2)
        self.assertRaises(AssertionError, x.cancel)

    def test_order(self):
        l = []
        fun = lambda x: l.append(x)
        for x in [0.05, 0.04, 0.03, 0.02, 0.01]:
            asynclib.call_later(x, fun, x)
        self.scheduler()
        self.assertEqual(l, [0.01, 0.02, 0.03, 0.04, 0.05])

    def test_delay(self):
        l = []
        fun = lambda x: l.append(x)
        asynclib.call_later(0.01, fun, 0.01).delay(0.07)
        asynclib.call_later(0.02, fun, 0.02).delay(0.08)
        asynclib.call_later(0.03, fun, 0.03)
        asynclib.call_later(0.04, fun, 0.04)
        asynclib.call_later(0.05, fun, 0.05)
        asynclib.call_later(0.06, fun, 0.06).delay(0.001)
        self.scheduler()
        self.assertEqual(l, [0.06, 0.03, 0.04, 0.05, 0.01, 0.02])

    def test_reset(self):
        l = []
        fun = lambda x: l.append(x)
        asynclib.call_later(0.01, fun, 0.01)
        asynclib.call_later(0.02, fun, 0.02)
        asynclib.call_later(0.03, fun, 0.03)
        x = asynclib.call_later(0.04, fun, 0.04)
        asynclib.call_later(0.05, fun, 0.05)
        time.sleep(0.1)
        x.reset()
        self.scheduler()
        self.assertEqual(l, [0.01, 0.02, 0.03, 0.05, 0.04])

    def test_cancel(self):
        l = []
        fun = lambda x: l.append(x)
        asynclib.call_later(0.01, fun, 0.01).cancel()
        asynclib.call_later(0.02, fun, 0.02)
        asynclib.call_later(0.03, fun, 0.03)
        asynclib.call_later(0.04, fun, 0.04)
        asynclib.call_later(0.05, fun, 0.05).cancel()
        self.scheduler()
        self.assertEqual(l, [0.02, 0.03, 0.04])



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

import unittest
from astm import codec
from astm.constants import STX, ETX, ETB, CR, LF, CRLF

def f(s, e='latin-1'):
    return s.format(STX=str(STX),
                       ETX=str(ETX),
                       ETB=str(ETB),
                       CR=str(CR),
                       LF=str(LF),
                       CRLF=str(CRLF)).encode(e)

class DecodeTestCase(unittest.TestCase):

    def test_decode_message(self):
        msg = f('{STX}1A|B|C|D{CR}{ETX}BF{CRLF}')
        res = [['A', 'B', 'C', 'D']]
        self.assertEqual(res, codec.decode(msg))

    def test_decode_message_with_nonascii(self):
        msg = f('{STX}1Й|Ц|У|К{CR}{ETX}F1{CRLF}', 'cp1251')
        res = [[u('Й'), u('Ц'), u('У'), u('К')]]
        self.assertEqual(res, codec.decode(msg, 'cp1251'))

    def test_decode_frame(self):
        msg = f('1A|B|C|D{CR}{ETX}')
        res = [['A', 'B', 'C', 'D']]
        self.assertEqual(res, codec.decode(msg))

    def test_decode_record(self):
        msg = f('A|B|C|D')
        res = [['A', 'B', 'C', 'D']]
        self.assertEqual(res, codec.decode(msg))


class DecodeMessageTestCase(unittest.TestCase):

    def test_decode_message(self):
        msg = f('{STX}1A|B|C|D{CR}{ETX}BF{CRLF}')
        res = (1, [['A', 'B', 'C', 'D']], 'BF')
        self.assertEqual(res, codec.decode_message(msg, 'ascii'))

    def test_fail_decome_message_with_wrong_checksumm(self):
        msg = f('{STX}1A|B|C|D{CR}{ETX}00{CRLF}')
        self.assertRaises(AssertionError, codec.decode_message, msg, 'ascii')

    def test_fail_decode_invalid_message(self):
        msg = f('A|B|C|D')
        self.assertRaises(ValueError, codec.decode_message, msg, 'ascii')

        msg = f('{STX}1A|B|C|D{CR}{ETX}BF')
        self.assertRaises(ValueError, codec.decode_message, msg, 'ascii')

        msg = f('1A|B|C|D{CR}{ETX}BF{CRLF}')
        self.assertRaises(ValueError, codec.decode_message, msg, 'ascii')


class DecodeFrameTestCase(unittest.TestCase):

    def test_fail_decode_without_tail_data(self):
        msg = f('1A|B|C')
        self.assertRaises(ValueError, codec.decode_frame, msg, 'ascii')

    def test_fail_decode_without_seq_value(self):
        msg = f('A|B|C|D{CR}{ETX}')
        self.assertRaises(ValueError, codec.decode_frame, msg, 'ascii')

    def test_fail_decode_with_invalid_tail(self):
        msg = f('1A|B|C{ETX}')
        self.assertRaises(ValueError, codec.decode_frame, msg, 'ascii')


class DecodeRecordTestCase(unittest.TestCase):

    def test_decode(self):
        msg = f('P|1|2776833|||ABC||||||||||||||||||||')
        res = ['P', '1', '2776833', None, None, 'ABC'] + [None] * 20
        self.assertEqual(res, codec.decode_record(msg, 'ascii'))

    def test_decode_with_components(self):
        msg = f('A|B^C^D^E|F')
        res = ['A', ['B', 'C', 'D', 'E'], 'F']
        self.assertEqual(res, codec.decode_record(msg, 'ascii'))

    def test_decode_with_repeated_components(self):
        msg = f(r'A|B^C\D^E|F')
        res = ['A', [['B', 'C'], ['D', 'E']], 'F']
        self.assertEqual(res, codec.decode_record(msg, 'ascii'))

    def test_decode_none_values_for_missed_ones(self):
        msg = f('A|||B')
        res = ['A', None, None, 'B']
        self.assertEqual(res, codec.decode_record(msg, 'ascii'))

        msg = f('A|B^^C^D^^E|F')
        res = ['A', ['B', None, 'C', 'D', None, 'E'], 'F']
        self.assertEqual(res, codec.decode_record(msg, 'ascii'))

    def test_decode_nonascii_chars_as_unicode(self):
        msg = f('привет|мир|!', 'utf8')
        res = [u('привет'), u('мир'), '!']
        self.assertEqual(res, codec.decode_record(msg, 'utf8'))


class EncodeTestCase(unittest.TestCase):

    def test_encode(self):
        msg = f('{STX}1A|B|C|D{CR}{ETX}BF{CRLF}')
        seq, data, cs = codec.decode_message(msg, 'ascii')
        self.assertEqual([msg], codec.encode(data, 'ascii'))

    def test_encode_message(self):
        msg = f('{STX}1A|B|C|D{CR}{ETX}BF{CRLF}')
        seq, data, cs = codec.decode_message(msg, 'ascii')
        self.assertEqual(msg, codec.encode_message(seq, data, 'ascii'))

    def test_encode_record(self):
        msg = br'A|B^C\D^E|F^G|H'
        record = codec.decode_record(msg, 'ascii')
        self.assertEqual(msg, codec.encode_record(record, 'ascii'))

    def test_encode_record_with_none_and_non_string(self):
        msg = ['foo', None, 0]
        res = b'foo||0'
        self.assertEqual(res, codec.encode_record(msg, 'ascii'))

    def test_encode_component(self):
        msg = ['foo', None, 0]
        res = b'foo^^0'
        self.assertEqual(res, codec.encode_component(msg, 'ascii'))

    def test_encode_component_strip_tail(self):
        msg = ['A', 'B', '', None, '']
        res = b'A^B'
        self.assertEqual(res, codec.encode_component(msg, 'ascii'))

    def test_encode_repeated_component(self):
        msg = [['foo', 1], ['bar', 2], ['baz', 3]]
        res = b'foo^1\\bar^2\\baz^3'
        self.assertEqual(res, codec.encode_repeated_component(msg, 'ascii'))

    def test_count_none_fields_as_empty_strings(self):
        self.assertEqual(b'|B|', codec.encode_record([None,'B', None], 'ascii'))

    def test_iter_encoding(self):
        records = [['foo', 1], ['bar', 2], ['baz', 3]]
        res = [f('{STX}1foo|1{CR}{ETX}32{CRLF}'),
               f('{STX}2bar|2{CR}{ETX}25{CRLF}'),
               f('{STX}3baz|3{CR}{ETX}2F{CRLF}')]
        self.assertEqual(res, list(codec.iter_encode(records, 'ascii')))

    def test_frame_number(self):
        records = list(map(list, 'ABCDEFGHIJ'))
        res = [f('{STX}1A{CR}{ETX}82{CRLF}'),
               f('{STX}2B{CR}{ETX}84{CRLF}'),
               f('{STX}3C{CR}{ETX}86{CRLF}'),
               f('{STX}4D{CR}{ETX}88{CRLF}'),
               f('{STX}5E{CR}{ETX}8A{CRLF}'),
               f('{STX}6F{CR}{ETX}8C{CRLF}'),
               f('{STX}7G{CR}{ETX}8E{CRLF}'),
               f('{STX}0H{CR}{ETX}88{CRLF}'),
               f('{STX}1I{CR}{ETX}8A{CRLF}'),
               f('{STX}2J{CR}{ETX}8C{CRLF}')]
        self.assertEqual(res, list(codec.iter_encode(records, 'ascii')))


class ChunkedEncodingTestCase(unittest.TestCase):

    def test_encode_chunky(self):
        recs = [['foo', 1], ['bar', 24], ['baz', [1, 2, 3], 'boo']]
        res = codec.encode(recs, size=14)
        self.assertTrue(isinstance(res, list))
        self.assertEqual(len(res), 4)

        self.assertEqual(res[0], f('{STX}1foo|1{CR}b{ETB}A8{CRLF}'))
        self.assertEqual(len(res[0]), 14)
        self.assertEqual(res[1], f('{STX}2ar|24{CR}b{ETB}6D{CRLF}'))
        self.assertEqual(len(res[1]), 14)
        self.assertEqual(res[2], f('{STX}3az|1^2^{ETB}C0{CRLF}'))
        self.assertEqual(len(res[2]), 14)
        self.assertEqual(res[3], f('{STX}43|boo{CR}{ETX}33{CRLF}'))
        self.assertLessEqual(len(res[3]), 14)


    def test_decode_chunks(self):
        recs = [['foo', 1], ['bar', 24], ['baz', [1, 2, 3], 'boo']]
        res = codec.encode(recs, size=14)
        for item in res:
            codec.decode(item)

    def test_join_chunks(self):
        recs = [['foo', '1'], ['bar', '24'], ['baz', ['1', '2', '3'], 'boo']]
        chunks = codec.encode(recs, size=14)
        msg = codec.join(chunks)
        self.assertEqual(codec.decode(msg), recs)

    def test_encode_as_single_message(self):
        res = codec.encode_message(2, [['A', 0]], 'ascii')
        self.assertEqual(f('{STX}2A|0{CR}{ETX}2F{CRLF}'), res)

    def test_is_chunked_message(self):
        msg = f('{STX}2A|0{CR}{ETB}2F{CRLF}')
        self.assertTrue(codec.is_chunked_message(msg))

        msg = f('{STX}2A|0{CR}{ETX}2F{CRLF}')
        self.assertFalse(codec.is_chunked_message(msg))


class ChecksummTestCase(unittest.TestCase):

    def test_common(self):
        msg = '2P|1|2776833|||王^大^明||||||||||||||||||||\x0D\x03'
        self.assertEqual(b'CF', codec.make_checksum(msg))

    def test_bytes(self):
        msg = '2P|1|2776833|||王^大^明||||||||||||||||||||\x0D\x03'.encode('utf8')
        self.assertEqual(b'4B', codec.make_checksum(msg))

    def test_short(self):
        self.assertEqual(b'02', codec.make_checksum('\x02'))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

try:
    from collections.abc import Iterable
except ImportError:  # Python 2
    from collections import Iterable
from .constants import (
    STX, ETX, ETB, CR, LF, CRLF,
    FIELD_SEP, COMPONENT_SEP, RECORD_SEP, REPEAT_SEP, ENCODING
)
try:
    from itertools import izip_longest
except ImportError:  # Python 3
    from itertools import zip_longest as izip_longest
import logging

log = logging.getLogger(__name__)


def decode(data, encoding=ENCODING):
    """Common ASTM decoding function that tries to guess which kind of data it
    handles.

    If `data` starts with STX character (``0x02``) than probably it is
    full ASTM message with checksum and other system characters.

    If `data` starts with digit character (``0-9``) than probably it is
    frame of records leading by his sequence number. No checksum is expected
    in this case.

    Otherwise it counts `data` as regular record structure.

    Note, that `data` should be bytes, not unicode string even if you know his
    `encoding`.

    :param data: ASTM data object.
    :type data: bytes

    :param encoding: Data encoding.
    :type encoding: str

    :return: List of ASTM records with unicode data.
    :rtype: list
    """
    if not isinstance(data, bytes):
        raise TypeError('bytes expected, got %r' % data)
    if data.startswith(STX):  # may be decode message \x02...\x03CS\r\n
        seq, records, cs = decode_message(data, encoding)
        return records
    byte = data[:1].decode()
    if  byte.isdigit():
        seq, records = decode_frame(data, encoding)
        return records
    return [decode_record(data, encoding)]


def decode_message(message, encoding):
    """Decodes complete ASTM message that is sent or received due
    communication routines. It should contains checksum that would be
    additionally verified.

    :param message: ASTM message.
    :type message: bytes

    :param encoding: Data encoding.
    :type encoding: str

    :returns: Tuple of three elements:

        * :class:`int` frame sequence number.
        * :class:`list` of records with unicode data.
        * :class:`bytes` checksum.

    :raises:
        * :exc:`ValueError` if ASTM message is malformed.
        * :exc:`AssertionError` if checksum verification fails.
    """
    if not isinstance(message, bytes):
        raise TypeError('bytes expected, got %r' % message)

    if not message.startswith(STX):
        raise ValueError('Malformed ASTM message: Must start with STX.')

    # Find the last ETX or ETB, as this marks the end of the frame data
    frame_end_pos = message.rfind(ETX)
    if frame_end_pos == -1:
        frame_end_pos = message.rfind(ETB)

    if frame_end_pos == -1:
        raise ValueError('Malformed ASTM message: No ETX or ETB found.')

    # The frame content for checksum includes the end character (ETX or ETB)
    frame_for_checksum = message[1:frame_end_pos + 1]

    # The trailer contains the checksum and final CR/LF
    trailer = message[frame_end_pos + 1:]
    cs = trailer.rstrip(CRLF)

    # Verify checksum
    ccs = make_checksum(frame_for_checksum)
    if cs != ccs:
        log.warning('Checksum failure: expected %r, calculated %r. Data may be corrupt.',
                    cs, ccs)

    # The actual records are in the frame *before* the ETX/ETB
    frame_content = frame_for_checksum[:-1]
    seq, records = decode_frame(frame_content, encoding)

    return seq, records, cs.decode('ascii') if cs else None


def decode_frame(frame, encoding):
    """Decodes ASTM frame: list of records optionally preceded by a sequence number."""
    if not isinstance(frame, bytes):
        raise TypeError('bytes expected, got %r' % frame)

    seq_byte = frame[:1]
    if seq_byte.isdigit():
        seq = int(seq_byte.decode('ascii'))
        records_data = frame[1:]
    else:
        seq = None
        records_data = frame

    # Records are separated by RECORD_SEP (\r).
    # A trailing separator will produce an empty string, which we should ignore.
    return seq, [decode_record(record, encoding)
                 for record in records_data.split(RECORD_SEP) if record]


def decode_record(record, encoding):
    """Decodes ASTM record message."""
    fields = []
    for item in record.split(FIELD_SEP):
        if REPEAT_SEP in item:
            item = decode_repeated_component(item, encoding)
        elif COMPONENT_SEP in item:
            item = decode_component(item, encoding)
        else:
            item = item.decode(encoding)
        fields.append([None, item][bool(item)])
    return fields


def decode_component(field, encoding):
    """Decodes ASTM field component."""
    return [[None, item.decode(encoding)][bool(item)]
            for item in field.split(COMPONENT_SEP)]


def decode_repeated_component(component, encoding):
    """Decodes ASTM field repeated component."""
    return [decode_component(item, encoding)
            for item in component.split(REPEAT_SEP)]


def encode(records, encoding=ENCODING, size=None, seq=1):
    """Encodes list of records into single ASTM message, also called as "packed"
    message.

    If you need to get each record as standalone message use :func:`iter_encode`
    instead.

    If the result message is too large (greater than specified `size` if it's
    not :const:`None`), than it will be split by chunks.

    :param records: List of ASTM records.
    :type records: list

    :param encoding: Data encoding.
    :type encoding: str

    :param size: Chunk size in bytes.
    :type size: int

    :param seq: Frame start sequence number.
    :type seq: int

    :return: List of ASTM message chunks.
    :rtype: list
    """
    msg = encode_message(seq, records, encoding)
    if size is not None and len(msg) > size:
        return list(split(msg, size))
    return [msg]


def iter_encode(records, encoding=ENCODING, size=None, seq=1):
    """Encodes and emits each record as separate message.

    If the result message is too large (greater than specified `size` if it's
    not :const:`None`), than it will be split by chunks.

    :yields: ASTM message chunks.
    :rtype: str
    """
    for record in records:
        msg = encode_message(seq, [record], encoding)
        if size is not None and len(msg) > size:
            for chunk in split(msg, size):
                seq += 1
                yield chunk
        else:
            seq += 1
            yield msg


def encode_message(seq, records, encoding):
    """Encodes ASTM message.

    :param seq: Frame sequence number.
    :type seq: int

    :param records: List of ASTM records.
    :type records: list

    :param encoding: Data encoding.
    :type encoding: str

    :return: ASTM complete message with checksum and other control characters.
    :rtype: str
    """
    data = RECORD_SEP.join(encode_record(record, encoding)
                           for record in records)
    data = b''.join((str(seq % 8).encode(), data, CR, ETX))
    return b''.join([STX, data, make_checksum(data), CR, LF])


def encode_record(record, encoding):
    """Encodes single ASTM record.

    :param record: ASTM record. Each :class:`str`-typed item counted as field
                   value, one level nested :class:`list` counted as components
                   and second leveled - as repeated components.
    :type record: list

    :param encoding: Data encoding.
    :type encoding: str

    :returns: Encoded ASTM record.
    :rtype: str
    """
    fields = []
    _append = fields.append
    for field in record:
        if isinstance(field, bytes):
            _append(field)
        elif isinstance(field, Iterable):
            _append(encode_component(field, encoding))
        elif field is None:
            _append(b'')
        else:
            _append(field.encode(encoding))
    return FIELD_SEP.join(fields)


def encode_component(component, encoding):
    """Encodes ASTM record field components."""
    items = []
    _append = items.append
    for item in component:
        if isinstance(item, bytes):
            _append(item)
        elif isinstance(item, Iterable):
            return encode_repeated_component(component, encoding)
        elif item is None:
            _append(b'')
        else:
            _append(item.encode(encoding))

    return COMPONENT_SEP.join(items).rstrip(COMPONENT_SEP)


def encode_repeated_component(components, encoding):
    """Encodes repeated components."""
    return REPEAT_SEP.join(encode_component(item, encoding)
                           for item in components)


def make_checksum(message):
    """Calculates checksum for specified message.

    :param message: ASTM message.
    :type message: bytes

    :returns: Checksum value that is actually byte sized integer in hex base
    :rtype: bytes
    """
    if not isinstance(message[0], int):
        message = map(ord, message)
    return hex(sum(message) & 0xFF)[2:].upper().zfill(2).encode()


def make_chunks(s, n):
    iter_bytes = (s[i:i + 1] for i in range(len(s)))
    return [b''.join(item)
            for item in izip_longest(*[iter_bytes] * n, fillvalue=b'')]


def split(msg, size):
    """Split `msg` into chunks with specified `size`.

    Chunk `size` value couldn't be less then 7 since each chunk goes with at
    least 7 special characters: STX, frame number, ETX or ETB, checksum and
    message terminator.

    :param msg: ASTM message.
    :type msg: bytes

    :param size: Chunk size in bytes.
    :type size: int

    :yield: `bytes`
    """
    stx, frame, msg, tail = msg[:1], msg[1:2], msg[2:-6], msg[-6:]
    assert stx == STX
    assert frame.isdigit()
    assert tail.endswith(CRLF)
    assert size is not None and size >= 7
    frame = int(frame)
    chunks = make_chunks(msg, size - 7)
    chunks, last = chunks[:-1], chunks[-1]
    idx = 0
    for idx, chunk in enumerate(chunks):
        item = b''.join([str((idx + frame) % 8).encode(), chunk, ETB])
        yield b''.join([STX, item, make_checksum(item), CRLF])
    item = b''.join([str((idx + frame + 1) % 8).encode(), last, CR, ETX])
    yield b''.join([STX, item, make_checksum(item), CRLF])


def join(chunks):
    """Merges ASTM message `chunks` into single message.

    :param chunks: List of chunks as `bytes`.
    :type chunks: iterable
    """
    msg = b'1' + b''.join(c[2:-5] for c in chunks) + ETX
    return b''.join([STX, msg, make_checksum(msg), CRLF])


def is_chunked_message(message):
    """Checks plain message for chunked byte."""
    length = len(message)
    if len(message) < 5:
        return False
    if ETB not in message:
        return False
    return message.index(ETB) == length - 5



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

import datetime
import unittest
from astm import codec
from astm import omnilab
from astm.constants import ENCODING

decode_record = lambda r: codec.decode_record(r.encode(), ENCODING)
encode_record = lambda r: codec.encode_record(r, ENCODING)


class OmnilabTestCase(unittest.TestCase):

    def test_header_request(self):
        data = r'H|\^&|||HOST^1.0.0|||||||P|E 1394-97|20091116104731'
        header = omnilab.client.Header(*decode_record(data))
        self.assertEqual(header.type, 'H')
        self.assertEqual(header.sender.name, 'HOST')
        self.assertEqual(header.sender.version, '1.0.0')
        self.assertEqual(header.version, 'E 1394-97')
        self.assertEqual(header.timestamp,
                         datetime.datetime(2009, 11, 16, 10, 47, 31))

    def test_header_response(self):
        data = r'H|\^&|||LabOnline^1.0.0|||||||P|E 1394-97|20091116104731'
        header = omnilab.server.Header(*decode_record(data))
        self.assertEqual(header.type, 'H')
        self.assertEqual(header.sender.name, 'LabOnline')
        self.assertEqual(header.sender.version, '1.0.0')
        self.assertEqual(header.version, 'E 1394-97')
        self.assertEqual(header.timestamp,
                         datetime.datetime(2009, 11, 16, 10, 47, 31))

    def test_patient_request(self):
        data = 'P|1|1212000|117118112||White^Nicky||19601218|M|||||Smith|37^years|0||||||||||CHIR'
        patient = omnilab.client.Patient(*decode_record(data))
        self.assertEqual(patient.type, 'P')
        self.assertEqual(patient.seq, 1)
        self.assertEqual(patient.practice_id, '1212000')
        self.assertEqual(patient.laboratory_id, '117118112')
        self.assertEqual(patient.name.last, 'White')
        self.assertEqual(patient.name.first, 'Nicky')
        self.assertEqual(patient.birthdate, datetime.datetime(1960, 12, 18))
        self.assertEqual(patient.sex, 'M')
        self.assertEqual(patient.physician_id, 'Smith')
        self.assertEqual(patient.special_1.value, 37)
        self.assertEqual(patient.special_1.unit, 'years')
        self.assertEqual(patient.special_2, 0)
        self.assertEqual(patient.location, 'CHIR')

    def test_patient_response(self):
        data = 'P|1|12120001|117118112||White^Nicky||19601218|M|||||||0||||||||||CHIR'
        patient = omnilab.server.Patient(*decode_record(data))
        self.assertEqual(patient.type, 'P')
        self.assertEqual(patient.seq, 1)
        self.assertEqual(patient.practice_id, '12120001')
        self.assertEqual(patient.laboratory_id, '117118112')
        self.assertEqual(patient.name.last, 'White')
        self.assertEqual(patient.name.first, 'Nicky')
        self.assertEqual(patient.birthdate, datetime.datetime(1960, 12, 18))
        self.assertEqual(patient.sex, 'M')
        self.assertEqual(patient.special_2, 0)
        self.assertEqual(patient.location, 'CHIR')

    def test_order_request(self):
        data = r'O|1|12120001||^^^NA^Sodium\^^^Cl^Clorum|R|20011023105715|20011023105715||||N||||S|||CHIM|AXM|Lab1|12120||||O|||||LAB2'
        order = omnilab.client.Order(*decode_record(data))
        self.assertEqual(order.type, 'O')
        self.assertEqual(order.seq, 1)
        self.assertEqual(order.sample_id, '12120001')
        self.assertEqual(order.test[0].assay_code, 'NA')
        self.assertEqual(order.test[0].assay_name, 'Sodium')
        self.assertEqual(order.test[1].assay_code, 'Cl')
        self.assertEqual(order.test[1].assay_name, 'Clorum')
        self.assertEqual(order.priority, 'R')
        self.assertEqual(order.created_at,
                         datetime.datetime(2001, 10, 23, 10, 57, 15))
        self.assertEqual(order.sampled_at,
                         datetime.datetime(2001, 10, 23, 10, 57, 15))
        self.assertEqual(order.action_code, 'N')
        self.assertEqual(order.biomaterial, 'S')
        self.assertEqual(order.user_field_1, 'CHIM')
        self.assertEqual(order.user_field_2, 'AXM')
        self.assertEqual(order.laboratory_field_1, 'Lab1')
        self.assertEqual(order.laboratory_field_2, '12120')
        self.assertEqual(order.report_type, 'O')
        self.assertEqual(order.laboratory, 'LAB2')

    def test_order_response(self):
        data = 'O|1|25140008|^1003^3|^^^Na^Sodium|R||||||||||U|||CHIM|ARCH||251400||||F'
        order = omnilab.server.Order(*decode_record(data))
        self.assertEqual(order.type, 'O')
        self.assertEqual(order.seq, 1)
        self.assertEqual(order.sample_id, '25140008')
        self.assertEqual(order.instrument.rack, '1003')
        self.assertEqual(order.instrument.position, '3')
        self.assertEqual(order.test.assay_code, 'Na')
        self.assertEqual(order.test.assay_name, 'Sodium')
        self.assertEqual(order.priority, 'R')
        self.assertEqual(order.created_at, None)
        self.assertEqual(order.sampled_at, None)
        self.assertEqual(order.action_code, None)
        self.assertEqual(order.biomaterial, 'U')
        self.assertEqual(order.user_field_1, 'CHIM')
        self.assertEqual(order.user_field_2, 'ARCH')
        self.assertEqual(order.laboratory_field_1, None)
        self.assertEqual(order.laboratory_field_2, '251400')
        self.assertEqual(order.report_type, 'F')
        self.assertEqual(order.laboratory, None)

    def test_result_request(self):
        data = 'R|1|^^^NA^Sodium|7.273|||||||||20091116104722|'
        result = omnilab.client.Result(*decode_record(data))
        self.assertEqual(result.type, 'R')
        self.assertEqual(result.seq, 1)
        self.assertEqual(result.test.assay_code, 'NA')
        self.assertEqual(result.test.assay_name, 'Sodium')
        self.assertEqual(result.value, '7.273')
        self.assertEqual(result.completed_at,
                         datetime.datetime(2009, 11, 16, 10, 47, 22))

    def test_result_response(self):
        data = 'R|1|^^^NA^Sodium|7.273|mmol/l|10-120|0|N|F||Val.Autom.^Smith |201009261006|201009261034^201009261033|Architect'
        result = omnilab.server.Result(*decode_record(data))
        self.assertEqual(result.type, 'R')
        self.assertEqual(result.seq, 1)
        self.assertEqual(result.test.assay_code, 'NA')
        self.assertEqual(result.test.assay_name, 'Sodium')
        self.assertEqual(result.units, 'mmol/l')
        self.assertEqual(result.references, '10-120')
        self.assertEqual(result.abnormal_flag, 0)
        self.assertEqual(result.abnormality_nature, 'N')
        self.assertEqual(result.status, 'F')
        self.assertEqual(result.operator.code_on_labonline, 'Val.Autom.')
        self.assertEqual(result.operator.code_on_analyzer, 'Smith ')
        self.assertEqual(result.started_at,
                         datetime.datetime(2010, 9, 26, 10, 0, 6))
        self.assertEqual(result.completed_at.labonline,
                         datetime.datetime(2010, 9, 26, 10, 3, 4))
        self.assertEqual(result.completed_at.analyzer,
                         datetime.datetime(2010, 9, 26, 10, 3, 3))
        self.assertEqual(result.instrument, 'Architect')

    def test_comment_request(self):
        data = 'C|1|L|SC^fully assured|G'
        comment = omnilab.client.Comment(*decode_record(data))
        self.assertEqual(comment.source, 'L')
        self.assertEqual(comment.data.code, 'SC')
        self.assertEqual(comment.data.value, 'fully assured')
        self.assertEqual(comment.ctype, 'G')

        data = 'C|1|I|TC^test comment|G'
        comment = omnilab.server.Comment(*decode_record(data))
        self.assertEqual(comment.source, 'I')
        self.assertEqual(comment.data.code, 'TC')
        self.assertEqual(comment.data.value, 'test comment')
        self.assertEqual(comment.ctype, 'G')

    def test_comment_response(self):
        data = 'C|1|I|SC^Sample contaminated|G'
        comment = omnilab.server.Comment(*decode_record(data))
        self.assertEqual(comment.source, 'I')
        self.assertEqual(comment.data.code, 'SC')
        self.assertEqual(comment.data.value, 'Sample contaminated')
        self.assertEqual(comment.ctype, 'G')

    def test_terminator_request(self):
        data = 'L|1|N'
        term = omnilab.client.Terminator(*decode_record(data))
        self.assertEqual(term.type, 'L')
        self.assertEqual(term.seq, 1)
        self.assertEqual(term.code, 'N')

    def test_terminator_response(self):
        data = 'L|1|N'
        term = omnilab.server.Terminator(*decode_record(data))
        self.assertEqual(term.type, 'L')
        self.assertEqual(term.seq, 1)
        self.assertEqual(term.code, 'N')

    def test_sample_check_in(self):
        data = 'C|1|I|CK^APS^20100925102955|G'
        comment = omnilab.server.Comment(*decode_record(data))
        self.assertEqual(comment.source, 'I')
        self.assertEqual(comment.data.code, 'CK')
        self.assertEqual(comment.data.value, 'APS')
        self.assertEqual(comment.data.field_1, '20100925102955')
        self.assertEqual(comment.ctype, 'G')

    def test_sample_seen(self):
        data = 'C|1|I|SE^APS^20100925102955|G'
        comment = omnilab.server.Comment(*decode_record(data))
        self.assertEqual(comment.source, 'I')
        self.assertEqual(comment.data.code, 'SE')
        self.assertEqual(comment.data.value, 'APS')
        self.assertEqual(comment.data.field_1, '20100925102955')
        self.assertEqual(comment.ctype, 'G')

    def test_force_close(self):
        data = 'C|1|I|CL|G'
        comment = omnilab.server.Comment(*decode_record(data))
        self.assertEqual(comment.source, 'I')
        self.assertEqual(comment.data.code, 'CL')
        self.assertEqual(comment.ctype, 'G')

    def test_add_test(self):
        data = 'C|1|I|TA|G'
        comment = omnilab.server.Comment(*decode_record(data))
        self.assertEqual(comment.source, 'I')
        self.assertEqual(comment.data.code, 'TA')
        self.assertEqual(comment.ctype, 'G')

    def test_sample_storage(self):
        data = 'C|1|I|SS^1^5^35^APS^20100925104523|G'
        comment = omnilab.server.Comment(*decode_record(data))
        self.assertEqual(comment.source, 'I')
        self.assertEqual(comment.data.code, 'SS')
        self.assertEqual(comment.data.value, '1')
        self.assertEqual(comment.data.field_1, '5')
        self.assertEqual(comment.data.field_2, '35')
        self.assertEqual(comment.data.field_3, 'APS')
        self.assertEqual(comment.data[-2], '20100925104523')
        self.assertEqual(comment.ctype, 'G')

    def test_host_query(self):
        data = 'C|1|I|HQ^CI8000_1^20100925103115|G'
        comment = omnilab.server.Comment(*decode_record(data))
        self.assertEqual(comment.source, 'I')
        self.assertEqual(comment.data.code, 'HQ')
        self.assertEqual(comment.data.value, 'CI8000_1')
        self.assertEqual(comment.data[-5], '20100925103115')
        self.assertEqual(comment.ctype, 'G')

    def test_aliquoting(self):
        data = 'C|1|I|AL^SERUM^500^56^25^fe500^20100925083115|G'
        comment = omnilab.server.Comment(*decode_record(data))
        self.assertEqual(comment.source, 'I')
        self.assertEqual(comment.data.code, 'AL')
        self.assertEqual(comment.data.value, 'SERUM')
        self.assertEqual(comment.data.field_1, '500')
        self.assertEqual(comment.data.field_2, '56')
        self.assertEqual(comment.data.field_3, '25')
        self.assertEqual(comment.data.field_4, 'fe500')
        self.assertEqual(comment.data.field_5, '20100925083115')
        self.assertEqual(comment.ctype, 'G')

    def test_primary_tube(self):
        data = 'C|1|I|PT^1^23^62^fe500^20100925083115|G'
        comment = omnilab.server.Comment(*decode_record(data))
        self.assertEqual(comment.source, 'I')
        self.assertEqual(comment.data.code, 'PT')
        self.assertEqual(comment.data.value, '1')
        self.assertEqual(comment.data.field_1, '23')
        self.assertEqual(comment.data.field_2, '62')
        self.assertEqual(comment.data.field_3, 'fe500')
        self.assertEqual(comment.data.field_4, '20100925083115')
        self.assertEqual(comment.ctype, 'G')
